from __future__ import annotations


__copyright__ = "Copyright (C) 2016 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from collections.abc import (
    Mapping as MappingABC,
    Set as abc_Set,
)
from dataclasses import dataclass
from functools import cached_property
from sys import intern
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Mapping,
    Sequence,
)
from warnings import warn

import islpy as isl
from pytools import ImmutableRecord, memoize_method
from pytools.tag import Tag, Taggable, tag_dataclass

from loopy.diagnostic import LoopyError
from loopy.tools import Optional as LoopyOptional


if TYPE_CHECKING:
    from loopy.types import LoopyType
    from loopy.typing import Expression, InameStr


# {{{ instruction tags

@tag_dataclass
class LegacyStringInstructionTag(Tag):
    """A subclass of :class:`pytools.tag.Tag` for use in
    :attr:`InstructionBase.tags` used for forward compatibility of the old
    string-based tagging mechanism. String-based tags are automatically converted
    to this type.

    .. attribute:: value
    """
    value: str

    # FIXME: This class should be deprecated as soon as there is a viable
    # alternative. For now, pattern matching and the textual syntax are
    # only able to generate string tags, which is why the deprecation is not
    # yet in effect.


@tag_dataclass
class UseStreamingStoreTag(Tag):
    """A subclass of :class:`pytools.tag.Tag` for use in
    :attr:`InstructionBase.tags` used to indicate that if the instruction is an
    :class:`Assignment` or a :class:`CallInstruction`, then the 'store' part of
    the assignment should be realized using streaming stores.

    .. note::

        This tag is advisory in nature and may be ignored by targets
        that do not understand it or in situations where it does not
        apply.

    .. warning::

        This is a dodgy shortcut, and no promise is made that this will
        continue to work. Whether this is safe is target-dependent and
        program-dependent. No promise of safety is made.
    """
    pass

# }}}


# {{{ HappensAfter

@dataclass(frozen=True)
class HappensAfter:
    """A class representing a "happens-after" relationship between two
    statements found in a :class:`loopy.LoopKernel`. Used to validate that a
    given kernel transformation respects the data dependencies in a given
    program.

    .. attribute:: variable_name

       The name of the variable responsible for the dependency. For
       backward compatibility purposes, this may be *None*. In this case, the
       dependency semantics revert to the deprecated, statement-level
       dependencies of prior versions of :mod:`loopy`.

    .. attribute:: instances_rel

        An :class:`islpy.Map` representing the precise happens-after
        relationship. The domain and range are sets of statement instances. The
        instances in the domain are required to execute before the instances in
        the range.

        Map dimensions are named according to the order of appearance of the
        inames in a :mod:`loopy` program. The dimension names in the range are
        appended with a prime to signify that the mapped instances are distinct.

        As a (deprecated) matter of backward compatibility, this may be *None*,
        in which case the semantics revert to the (underspecified)
        statement-level dependencies of prior versions of :mod:`loopy`.
    """

    variable_name: str | None
    instances_rel: isl.Map | None

# }}}


# {{{ instructions: base class

class InstructionBase(ImmutableRecord, Taggable):
    """A base class for all types of instruction that can occur in
    a kernel.

    .. attribute:: id

        An (otherwise meaningless) identifier that is unique within
        a :class:`loopy.LoopKernel`.

    .. rubric:: Instruction ordering

    .. attribute:: depends_on

        a :class:`frozenset` of :attr:`id` values of :class:`InstructionBase`
        instances that *must* be executed before this one. Note that
        :func:`loopy.preprocess_kernel` (usually invoked automatically)
        augments this by adding dependencies on any writes to temporaries read
        by this instruction.

        May be *None* to invoke the default.

        There are two extensions to this:

        - You may use `*` as a wildcard in the given IDs. This will be expanded
          to all matching instruction IDs during :func:`loopy.make_kernel`.
        - Instead of an instruction ID, you may pass an instance of
          :class:`loopy.match.MatchExpressionBase` into the :attr:`depends_on`
          :class:`frozenset`. The given expression will be used to add any
          matching instructions in the kernel to :attr:`depends_on` during
          :func:`loopy.make_kernel`. Note, that this is not meant as a user-facing
          interface.

    .. attribute:: depends_on_is_final

        A :class:`bool` determining whether :attr:`depends_on` constitutes
        the *entire* list of iname dependencies. If *not* marked final,
        various semi-broken heuristics will try to add further dependencies.

        Defaults to *False*.

    .. attribute:: groups

        A :class:`frozenset` of strings indicating the names of 'instruction
        groups' of which this instruction is a part. An instruction group is
        considered 'active' as long as one (but not all) instructions of the
        group have been executed.

    .. attribute:: conflicts_with_groups

        A :class:`frozenset` of strings indicating which instruction groups
        (see :attr:`groups`) may not be active when this
        instruction is scheduled.

    .. attribute:: priority

        Scheduling priority, an integer. Higher means 'execute sooner'.
        Default 0.

    .. rubric :: Synchronization

    .. attribute:: no_sync_with

        a :class:`frozenset` of tuples of the form ``(insn_id, scope)``, where
        ``insn_id`` refers to :attr:`id` of :class:`InstructionBase` instances
        and `scope` is one of the following strings:

           - `"local"`
           - `"global"`
           - `"any"`.

        An element ``(insn_id, scope)`` means "do not consider any variable
        access conflicting for variables of ``scope`` between this instruction
        and ``insn_id``".
        Specifically, loopy will not complain even if it detects that accesses
        potentially requiring ordering (e.g. by dependencies) exist, and it
        will not emit barriers to guard any dependencies from this
        instruction on ``insn_id`` that may exist.

        Note, that :attr:`no_sync_with` allows instruction matching through wildcards
        and match expression, just like :attr:`depends_on`.

        This data is used specifically by barrier insertion and
        :func:`loopy.check.check_variable_access_ordered`.

    .. rubric:: Conditionals

    .. attribute:: predicates

        a :class:`frozenset` of expressions. The conjunction (logical and) of
        their truth values (as defined by C) determines whether this instruction
        should be run.

    .. rubric:: Iname dependencies

    .. attribute:: within_inames

        A :class:`frozenset` of inames identifying the loops within which this
        instruction will be executed.

    .. rubric:: Iname dependencies

    .. rubric:: Tagging

    .. attribute:: tags

        A :class:`frozenset` of subclasses of :class:`pytools.tag.Tag` used to
        provide metadata on this object. Legacy string tags are converted to
        :class:`LegacyStringInstructionTag` or, if they used to carry
        a functional meaning, the tag carrying that same functional meaning
        (e.g. :class:`UseStreamingStoreTag`).

    .. automethod:: __init__
    .. automethod:: assignee_var_names
    .. automethod:: assignee_subscript_deps
    .. automethod:: with_transformed_expressions
    .. automethod:: write_dependency_names
    .. automethod:: dependency_names
    .. automethod:: copy

    Inherits from :class:`pytools.tag.Taggable`.
    """
    id: str | None
    happens_after: Mapping[str, HappensAfter]
    depends_on_is_final: bool
    groups: frozenset[str]
    conflicts_with_groups: frozenset[str]
    no_sync_with: frozenset[tuple[str, str]]
    predicates: frozenset[Expression]
    within_inames: frozenset[InameStr]
    within_inames_is_final: bool
    priority: int

    # within_inames_is_final is deprecated and will be removed in version 2017.x.

    fields: ClassVar[set[str]] = set("id depends_on_is_final "
            "groups conflicts_with_groups "
            "no_sync_with "
            "predicates "
            "within_inames_is_final within_inames "
            "priority".split())

    def __init__(self,
                 id: str | None,
                 happens_after: (
                     Mapping[str, HappensAfter] | frozenset[str] | str | None),
                 depends_on_is_final: bool | None,
                 groups: frozenset[str] | None,
                 conflicts_with_groups: frozenset[str] | None,
                 no_sync_with: frozenset[tuple[str, str]] | None,
                 within_inames_is_final: bool | None,
                 within_inames: frozenset[str] | None,
                 priority: int | None,
                 predicates: frozenset[str] | None,
                 tags: frozenset[Tag] | None,
                 *,
                 depends_on: frozenset[str] | str | None = None,
                 ) -> None:
        from constantdict import constantdict

        if predicates is None:
            predicates = frozenset()

        new_predicates = set()
        for pred in predicates:
            if isinstance(pred, str):
                from loopy.symbolic import parse
                pred = parse(pred)

            new_predicates.add(pred)

        predicates = frozenset(new_predicates)
        del new_predicates

        # {{{ process happens_after/depends_on

        if happens_after is not None and depends_on is not None:
            raise TypeError("may not pass both happens_after and depends_on")
        elif depends_on is not None:
            # FIXME Enable once we realistically check detailed dependencies.
            # warn("depends_on is deprecated and will stop working in 2026. "
            #      "Pass happens_after instead.", DeprecationWarning, stacklevel=2)
            happens_after = depends_on

        del depends_on

        if depends_on_is_final and happens_after is None:
            raise LoopyError("Setting depends_on_is_final to True requires "
                    "actually specifying happens_after/depends_on")

        if isinstance(happens_after, constantdict):
            pass
        elif happens_after is None:
            happens_after = constantdict()
        elif isinstance(happens_after, str):
            warn("Passing a string for happens_after/depends_on is deprecated and "
                 "will stop working in 2025. Instead, pass a full-fledged "
                 "happens_after data structure.", DeprecationWarning, stacklevel=2)

            happens_after = constantdict({
                    after_id.strip(): HappensAfter(
                        variable_name=None,
                        instances_rel=None)
                    for after_id in happens_after.split(",")
                    if after_id.strip()})
        elif isinstance(happens_after, frozenset):
            happens_after = constantdict({
                    after_id: HappensAfter(
                        variable_name=None,
                        instances_rel=None)
                    for after_id in happens_after})
        elif isinstance(happens_after, dict):
            happens_after = constantdict(happens_after)
        else:
            raise TypeError("'happens_after' has unexpected type: "
                            f"{type(happens_after)}")

        # }}}

        if groups is None:
            groups = frozenset()

        if conflicts_with_groups is None:
            conflicts_with_groups = frozenset()

        if no_sync_with is None:
            no_sync_with = frozenset()

        if within_inames is None:
            within_inames = frozenset()

        if within_inames_is_final is None:
            within_inames_is_final = False

        if depends_on_is_final is None:
            depends_on_is_final = False

        if depends_on_is_final and not isinstance(happens_after, MappingABC):
            raise LoopyError("Setting depends_on_is_final to True requires "
                    "actually specifying happens_after/depends_on")

        if tags is None:
            tags = frozenset()

        if priority is None:
            priority = 0

        if not isinstance(tags, abc_Set):
            # was previously allowed to be tuple
            tags = frozenset(tags)

        # Periodically reenable these and run the tests to ensure all
        # performance-relevant identifiers are interned.
        #
        # from loopy.tools import is_interned
        # assert is_interned(id)
        # assert all(is_interned(dep) for dep in depends_on)
        # assert all(is_interned(grp) for grp in groups)
        # assert all(is_interned(grp) for grp in conflicts_with_groups)
        # assert all(is_interned(iname) for iname in within_inames)
        # assert all(is_interned(pred) for pred in predicates)

        assert isinstance(within_inames, abc_Set)
        assert isinstance(happens_after, MappingABC) or happens_after is None
        assert isinstance(groups, abc_Set)
        assert isinstance(conflicts_with_groups, abc_Set)

        from loopy.tools import is_hashable
        assert is_hashable(happens_after)

        ImmutableRecord.__init__(self,
                id=id,
                happens_after=happens_after,
                depends_on_is_final=depends_on_is_final,
                no_sync_with=no_sync_with,
                groups=groups, conflicts_with_groups=conflicts_with_groups,
                within_inames_is_final=within_inames_is_final,
                within_inames=within_inames,
                priority=priority,
                predicates=predicates,
                # Yes, tags is set by both this and the Taggable constructor.
                # Here, we set it so that ImmutableRecord knows about it.
                # The Taggable constructor call does extra validation.
                tags=tags)

    def get_copy_kwargs(self, **kwargs):
        passed_depends_on = "depends_on" in kwargs

        if passed_depends_on:
            assert "happens_after" not in kwargs

        kwargs = super().get_copy_kwargs(**kwargs)

        if passed_depends_on:
            # FIXME Enable once we realistically check detailed dependencies.
            # warn("depends_on is deprecated and will stop working in 2026. "
            #      "Instead, use happens_after.", DeprecationWarning, stacklevel=2)
            del kwargs["happens_after"]

        return kwargs

    # {{{ abstract interface

    def read_dependency_names(self):
        from loopy.symbolic import get_dependencies
        result = frozenset()

        for pred in self.predicates:
            result = result | get_dependencies(pred)

        return result

    def reduction_inames(self) -> frozenset[str]:
        raise NotImplementedError

    def sub_array_ref_inames(self):
        raise NotImplementedError

    def assignee_var_names(self):
        """Return a tuple of assignee variable names, one
        for each quantity being assigned to.
        """
        raise NotImplementedError

    def assignee_subscript_deps(self):
        """Return a list of sets of variable names referred to in the subscripts
        of the quantities being assigned to, one for each assignee.
        """
        raise NotImplementedError

    def with_transformed_expressions(self, f, assignee_f=None):
        """Return a new copy of *self* where *f* has been applied to every
        expression occurring in *self*. *args* will be passed as extra
        arguments (in addition to the expression) to *f*.

        If *assignee_f* is passed, then left-hand sides of assignments are
        passed to it. If it is not given, it defaults to the same as *f*.
        """
        raise NotImplementedError

    # }}}

    @property
    def depends_on(self):
        # FIXME Enable once we realistically check detailed dependencies.
        # warn("depends_on is deprecated and will stop working in 2026. "
        #      "Use happens_after instead.", DeprecationWarning, stacklevel=2)
        return frozenset(self.happens_after)

    @property
    def assignee_name(self):
        """A convenience wrapper around :meth:`assignee_var_names`
        that returns the the name of the variable being assigned.
        If more than one variable is being modified in the instruction,
        :raise:`ValueError` is raised.
        """

        names = self.assignee_var_names()

        if len(names) != 1:
            raise ValueError("expected exactly one assignment in instruction "
                    "on which assignee_name is being called, found %d"
                    % len(names))

        name, = names
        return name

    @memoize_method
    def write_dependency_names(self):
        """Return a set of dependencies of the left hand side of the
        assignments performed by this instruction, including written variables
        and indices.
        """

        result = frozenset(self.assignee_var_names())
        asd = self.assignee_subscript_deps()
        if asd:
            result = result | frozenset.union(*self.assignee_subscript_deps())

        return result

    @memoize_method
    def dependency_names(self):
        return self.read_dependency_names() | self.write_dependency_names()

    def get_str_options(self):
        result = []

        if self.depends_on:
            result.append("dep="+":".join(self.depends_on))
        if self.no_sync_with:
            result.append("nosync="+":".join(
                    "%s@%s" % entry for entry in self.no_sync_with))
        if self.groups:
            result.append("groups=%s" % ":".join(self.groups))
        if self.conflicts_with_groups:
            result.append("conflicts=%s" % ":".join(self.conflicts_with_groups))
        if self.priority:
            result.append("priority=%d" % self.priority)
        if self.tags:
            from loopy.kernel.tools import stringify_instruction_tag
            result.append("tags=%s" % ":".join(
                stringify_instruction_tag(t) for t in self.tags))
        if hasattr(self, "atomicity") and self.atomicity:
            result.append("atomic=%s" % ":".join(str(a) for a in self.atomicity))

        return result

    # {{{ hashing and key building

    @cached_property
    def _key_builder(self):
        from loopy.tools import LoopyEqKeyBuilder
        key_builder = LoopyEqKeyBuilder()
        key_builder.update_for_class(self.__class__)

        for field_name in self.fields:
            key_builder.update_for_field(field_name, getattr(self, field_name))

        return key_builder

    def update_persistent_hash(self, key_hash, key_builder):
        """Custom hash computation function for use with
        :class:`pytools.persistent_dict.PersistentDict`.

        Only works in conjunction with :class:`loopy.tools.KeyBuilder`.
        """

        key_builder.rec(key_hash, self._key_builder.hash_key())

    # }}}

    def __setstate__(self, val):
        super().__setstate__(val)

        from constantdict import constantdict

        from loopy.tools import intern_frozenset_of_ids

        if self.id is not None:  # pylint:disable=access-member-before-definition
            self.id = intern(self.id)
        self.happens_after = constantdict({
                intern(after_id): ha
                for after_id, ha in self.happens_after.items()})
        self.groups = intern_frozenset_of_ids(self.groups)
        self.conflicts_with_groups = (
                intern_frozenset_of_ids(self.conflicts_with_groups))
        self.within_inames = (
                intern_frozenset_of_ids(self.within_inames))

    def _with_new_tags(self, tags: frozenset[Tag]):
        return self.copy(tags=tags)

# }}}


def _get_assignee_var_name(expr):
    from pymbolic.primitives import Lookup, Subscript, Variable

    from loopy.symbolic import LinearSubscript, SubArrayRef

    if isinstance(expr, Lookup):
        expr = expr.aggregate

    if isinstance(expr, Variable):
        return expr.name

    elif isinstance(expr, Subscript):
        agg = expr.aggregate
        assert isinstance(agg, Variable)

        return agg.name

    elif isinstance(expr, LinearSubscript):
        agg = expr.aggregate
        assert isinstance(agg, Variable)

        return agg.name

    elif isinstance(expr, SubArrayRef):
        agg = expr.subscript.aggregate
        assert isinstance(agg, Variable)

        return agg.name

    else:
        raise RuntimeError("invalid lvalue '%s'" % expr)


def _get_assignee_subscript_deps(expr):
    from pymbolic.primitives import Lookup, Subscript, Variable

    from loopy.symbolic import LinearSubscript, SubArrayRef, get_dependencies

    if isinstance(expr, Lookup):
        expr = expr.aggregate

    if isinstance(expr, Variable):
        return frozenset()
    elif isinstance(expr, Subscript):
        return get_dependencies(expr.index)
    elif isinstance(expr, LinearSubscript):
        return get_dependencies(expr.index)
    elif isinstance(expr, SubArrayRef):
        return get_dependencies(expr.subscript.index) - (
                frozenset(iname.name for iname in expr.swept_inames))
    else:
        raise RuntimeError("invalid lvalue '%s'" % expr)


# {{{ atomic ops

class MemoryOrdering:
    """Ordering of atomic operations, defined as in C11 and OpenCL.

    .. attribute:: RELAXED
    .. attribute:: ACQUIRE
    .. attribute:: RELEASE
    .. attribute:: ACQ_REL
    .. attribute:: SEQ_CST
    """

    RELAXED = 0
    ACQUIRE = 1
    RELEASE = 2
    ACQ_REL = 3
    SEQ_CST = 4

    @staticmethod
    def to_string(v):
        for i in dir(MemoryOrdering):
            if i.startswith("_"):
                continue

            if getattr(MemoryOrdering, i) == v:
                return i

        raise ValueError("Unknown value of MemoryOrdering")


class MemoryScope:
    """Scope of atomicity, defined as in OpenCL.

    .. attribute:: auto

        Scope matches the accessibility of the variable.

    .. attribute:: WORK_ITEM
    .. attribute:: WORK_GROUP
    .. attribute:: WORK_DEVICE
    .. attribute:: ALL_SVM_DEVICES
    """

    WORK_ITEM = 0
    WORK_GROUP = 1
    DEVICE = 2
    ALL_SVM_DEVICES = 2

    auto = -1

    @staticmethod
    def to_string(v):
        for i in dir(MemoryScope):
            if i.startswith("_"):
                continue

            if getattr(MemoryScope, i) == v:
                return i

        raise ValueError("Unknown value of MemoryScope")


class VarAtomicity:
    """A base class for the description of how atomic access to :attr:`var_name`
    shall proceed.

    .. attribute:: var_name
    """

    def __init__(self, var_name):
        self.var_name = var_name

    def update_persistent_hash(self, key_hash, key_builder):
        """Custom hash computation function for use with
        :class:`pytools.persistent_dict.PersistentDict`.
        """

        key_builder.rec(key_hash, self.var_name)

    def __eq__(self, other):
        return (type(self) is type(other)
                and self.var_name == other.var_name)

    def __ne__(self, other):
        return not self.__eq__(other)


class OrderedAtomic(VarAtomicity):
    """Properties of an atomic operation. A subclass of :class:`VarAtomicity`.

    .. attribute:: ordering

        One of the values from :class:`MemoryOrdering`

    .. attribute:: scope

        One of the values from :class:`MemoryScope`
    """

    ordering = MemoryOrdering.SEQ_CST
    scope = MemoryScope.auto

    def update_persistent_hash(self, key_hash, key_builder):
        """Custom hash computation function for use with
        :class:`pytools.persistent_dict.PersistentDict`.
        """

        super().update_persistent_hash(key_hash, key_builder)
        key_builder.rec(key_hash, str(self.__class__.__name__))
        key_builder.rec(key_hash, self.ordering)
        key_builder.rec(key_hash, self.scope)

    def __eq__(self, other):
        return (super().__eq__(other)
                and self.ordering == other.ordering
                and self.scope == other.scope)

    def __hash__(self):
        return hash((type(self), self.var_name,
                     self.ordering, self.scope))

    @property
    def op_name(self):
        raise NotImplementedError

    def __str__(self):
        return "{}[{}]{}/{}".format(
                self.op_name,
                self.var_name,
                MemoryOrdering.to_string(self.ordering),
                MemoryScope.to_string(self.scope))


class AtomicInit(OrderedAtomic):
    """Describes initialization of an atomic variable. A subclass of
    :class:`OrderedAtomic`.

    .. attribute:: ordering

        One of the values from :class:`MemoryOrdering`

    .. attribute:: scope

        One of the values from :class:`MemoryScope`
    """
    op_name = "init"


class AtomicUpdate(OrderedAtomic):
    """Properties of an atomic update. A subclass of
    :class:`OrderedAtomic`.

    .. attribute:: ordering

        One of the values from :class:`MemoryOrdering`

    .. attribute:: scope

        One of the values from :class:`MemoryScope`
    """
    op_name = "update"


class AtomicLoad(OrderedAtomic):
    """Properties of an atomic load. A subclass of :class:`OrderedAtomic`.

    .. attribute:: ordering

        One of the values from :class:`MemoryOrdering`

    .. attribute:: scope

        One of the values from :class:`MemoryScope`
    """
    op_name = "load"

# }}}


# {{{ instruction base class: expression rhs

class MultiAssignmentBase(InstructionBase):
    """An assignment instruction with an expression as a right-hand side."""

    fields = InstructionBase.fields | {"expression"}

    @memoize_method
    def read_dependency_names(self):
        from loopy.symbolic import get_dependencies
        result = (
                super().read_dependency_names()
                | get_dependencies(self.expression))

        for subscript_deps in self.assignee_subscript_deps():
            result = result | subscript_deps

        return result

    @memoize_method
    def reduction_inames(self):
        from loopy.symbolic import get_reduction_inames
        return frozenset(get_reduction_inames(self.expression))

    @memoize_method
    def sub_array_ref_inames(self):
        from loopy.symbolic import get_sub_array_ref_swept_inames
        return get_sub_array_ref_swept_inames((self.assignees, self.expression))

# }}}


# {{{ instruction: assignment

class _not_provided:  # noqa: N801
    pass


class Assignment(MultiAssignmentBase):
    """
    .. attribute:: assignee

    .. attribute:: expression

    The following attributes are only used until
    :func:`loopy.make_kernel` is finished:

    .. attribute:: temp_var_type

        A :class:`loopy.Optional`. If not empty, contains the type that
        will be assigned to the new temporary variable created from the
        assignment.

    .. attribute:: atomicity

        A tuple of instances of :class:`VarAtomicity`. Together, they describe
        to what extent the assignment is to be carried out in a way that
        involves atomic operations.

        To describe an atomic update, any memory reads of *exact* occurrences
        of the left-hand side expression of the assignment in the right hand
        side are treated , together with the "memory write" part of the
        assignment, as part of one single atomic update.

        .. note::

            Exact identity of the LHS with RHS subexpressions is required for
            an atomic update to be recognized. For example, the following update
            will not be recognized as an update::

                z[i] = z[i+1-1] + a {atomic}

        :mod:`loopy` may choose to evaluate the right-hand side *multiple times*
        as part of a single assignment. It is up to the user to ensure that
        this retains correct semantics.

        For example, the following assignment::

            z[i] = f(z[i]) + a {atomic}

        may generate the following (pseudo-)code::

            DO
                READ ztemp_old = z[i]
                EVALUATE ztemp_new = f(ztemp_old) + a
            WHILE compare_and_swap(z[i], ztemp_new, ztemp_old) did not succeed

    .. automethod:: __init__
    """

    assignee: Expression
    expression: Expression
    temp_var_type: LoopyOptional
    atomicity: tuple[VarAtomicity, ...]

    fields = MultiAssignmentBase.fields | \
            set("assignee temp_var_type atomicity".split())

    def __init__(self,
                 assignee: str | Expression,
                 expression: str | Expression,
                 id: str | None = None,
                 happens_after:
                     Mapping[str, HappensAfter] | frozenset[str] | str | None = None,
                 depends_on_is_final: bool | None = None,
                 groups: frozenset[str] | None = None,
                 conflicts_with_groups: frozenset[str] | None = None,
                 no_sync_with: frozenset[tuple[str, str]] | None = None,
                 within_inames_is_final: bool | None = None,
                 within_inames: frozenset[str] | None = None,
                 priority: int | None = None,
                 predicates: frozenset[str] | None = None,
                 tags: frozenset[Tag] | None = None,
                 temp_var_type:
                    type[_not_provided] | LoopyOptional | LoopyType | None
                    = _not_provided,
                 atomicity: tuple[VarAtomicity, ...] = (),
                 *,
                 depends_on: frozenset[str] | str | None = None,
                 ) -> None:

        if temp_var_type is _not_provided:
            temp_var_type = LoopyOptional()

        super().__init__(
                id=id,
                happens_after=happens_after,
                depends_on_is_final=depends_on_is_final,
                groups=groups,
                conflicts_with_groups=conflicts_with_groups,
                no_sync_with=no_sync_with,
                within_inames_is_final=within_inames_is_final,
                within_inames=within_inames,
                priority=priority,
                predicates=predicates,
                tags=tags,
                depends_on=depends_on)

        from loopy.symbolic import parse
        if isinstance(assignee, str):
            assignee = parse(assignee)
        if isinstance(expression, str):
            parsed_expression = parse(expression)
        else:
            parsed_expression = expression

        from pymbolic.primitives import Lookup, Subscript, Variable

        from loopy.symbolic import LinearSubscript
        if not isinstance(assignee, (Variable, Subscript, LinearSubscript, Lookup)):
            raise LoopyError("invalid lvalue '%s'" % assignee)

        self.assignee = assignee
        self.expression = parsed_expression

        self.temp_var_type = _check_and_fix_temp_var_type(temp_var_type)
        self.atomicity = atomicity

    # {{{ implement InstructionBase interface

    @memoize_method
    def assignee_var_names(self):
        return (_get_assignee_var_name(self.assignee),)

    def assignee_subscript_deps(self):
        return (_get_assignee_subscript_deps(self.assignee),)

    def with_transformed_expressions(self, f, assignee_f=None):
        if assignee_f is None:
            assignee_f = f

        assignee = assignee_f(self.assignee)
        expression = f(self.expression)
        predicates = []
        changed_predicates = False
        for pred in self.predicates:
            new_pred = f(pred)
            if new_pred is not pred:
                changed_predicates = True
            predicates.append(new_pred)
        if changed_predicates:
            predicates = frozenset(predicates)
        else:
            predicates = self.predicates

        if assignee is self.assignee and expression is self.expression and \
                predicates is self.predicates:
            return self

        return self.copy(
                assignee=assignee,
                expression=expression,
                predicates=predicates)

    # }}}

    def __str__(self):
        result = f"{self.assignee} <- {self.expression}"

        if self.id is not None:
            result = "%s: " % self.id + result

        options = self.get_str_options()
        if options:
            result += " {%s}" % (": ".join(options))

        if self.predicates:
            result += "\n" + 10*" " + "if (%s)" % " and ".join(
                    str(p) for p in self.predicates)
        return result

    # {{{ for interface uniformity with CallInstruction

    @property
    def temp_var_types(self):
        return (self.temp_var_type,)

    @property
    def assignees(self):
        return (self.assignee,)

    @memoize_method
    def sub_array_ref_inames(self):
        assert super().sub_array_ref_inames() == frozenset()
        return frozenset()

    # }}}

# }}}


# {{{ instruction: function call

class CallInstruction(MultiAssignmentBase):
    """An instruction capturing a function call. Unlike :class:`Assignment`,
    this instruction supports functions with multiple return values.

    .. attribute:: assignees

        A :class:`tuple` of left-hand sides for the assignment

    .. attribute:: expression

    The following attributes are only used until
    :func:`loopy.make_kernel` is finished:

    .. attribute:: temp_var_types

        A tuple of `:class:loopy.Optional`. If an entry is not empty, it
        contains the type that will be assigned to the new temporary variable
        created from the assignment.

    .. automethod:: __init__
    """

    fields = MultiAssignmentBase.fields | \
            set("assignees temp_var_types".split())

    def __init__(self,
            assignees, expression,
            id=None,
            happens_after=None,
            depends_on_is_final=None,
            groups=None,
            conflicts_with_groups=None,
            no_sync_with=None,
            within_inames_is_final=None,
            within_inames=None,
            tags=None,
            temp_var_types=None,
            priority=0, predicates=frozenset(),
            depends_on=None):

        super().__init__(
                id=id,
                happens_after=happens_after,
                depends_on_is_final=depends_on_is_final,
                groups=groups,
                conflicts_with_groups=conflicts_with_groups,
                no_sync_with=no_sync_with,
                within_inames_is_final=within_inames_is_final,
                within_inames=within_inames,
                priority=priority,
                predicates=predicates,
                tags=tags,
                depends_on=depends_on)

        from pymbolic.primitives import Call

        from loopy.symbolic import Reduction
        if not isinstance(expression, (Call, Reduction)) and (
                expression is not None):
            raise LoopyError("'expression' argument to CallInstruction "
                    "must be a function call")

        from loopy.symbolic import parse
        if isinstance(assignees, str):
            assignees = parse(assignees)
        if not isinstance(assignees, tuple):
            raise LoopyError("'assignees' argument to CallInstruction "
                    "must be a tuple or a string parseable to a tuple"
                    "--got '%s'" % type(assignees).__name__)

        if isinstance(expression, str):
            expression = parse(expression)

        from pymbolic.primitives import Subscript, Variable

        from loopy.symbolic import LinearSubscript, SubArrayRef
        for assignee in assignees:
            if not isinstance(assignee, (Variable, Subscript, LinearSubscript,
                    SubArrayRef)):
                raise LoopyError("invalid lvalue '%s'" % assignee)

        self.assignees = assignees
        self.expression = expression

        if temp_var_types is None:
            self.temp_var_types = (LoopyOptional(),) * len(self.assignees)
        else:
            self.temp_var_types = tuple(
                    _check_and_fix_temp_var_type(tvt, stacklevel=3)
                    for tvt in temp_var_types)

    # {{{ implement InstructionBase interface

    @memoize_method
    def assignee_var_names(self):
        return tuple(_get_assignee_var_name(a) for a in self.assignees)

    def assignee_subscript_deps(self):
        return tuple(
                _get_assignee_subscript_deps(a)
                for a in self.assignees)

    def with_transformed_expressions(self, f, assignee_f=None):
        if assignee_f is None:
            assignee_f = f

        assignees = assignee_f(self.assignees)
        expression = f(self.expression)
        predicates = []
        changed_predicates = False
        for pred in self.predicates:
            new_pred = f(pred)
            if new_pred is not pred:
                changed_predicates = True
            predicates.append(new_pred)
        if changed_predicates:
            predicates = frozenset(predicates)
        else:
            predicates = self.predicates

        if len(assignees) == len(self.assignees) and \
                all(assignee is orig_assignee for assignee, orig_assignee in
                    zip(assignees, self.assignees)) \
                and expression is self.expression and \
                predicates is self.predicates:
            return self

        return self.copy(
                assignees=assignees,
                expression=expression,
                predicates=predicates)

    # }}}

    def __str__(self):
        result = "{}: {} <- {}".format(self.id,
                ", ".join(str(a) for a in self.assignees),
                self.expression)

        options = self.get_str_options()
        if options:
            result += " {%s}" % (": ".join(options))

        if self.predicates:
            result += "\n" + 10*" " + "if (%s)" % " && ".join(
                    str(pred) for pred in self.predicates)
        return result

    def arg_id_to_arg(self):
        """:returns: a :class:`dict` mapping argument identifiers (non-negative
            numbers for positional arguments and negative numbers for assignees) to
            their respective values
        """
        arg_id_to_arg = dict(enumerate(self.expression.parameters))
        for i, arg in enumerate(self.assignees):
            arg_id_to_arg[-i-1] = arg

        return arg_id_to_arg

    @property
    def atomicity(self):
        # Function calls can impossibly be atomic, and even the result assignment
        # is troublesome, especially in the case of multiple results. Avoid the
        # issue altogether by disallowing atomicity.
        return ()

# }}}


def subscript_contains_slice(subscript):
    """Return *True* if the *subscript* contains an instance of
    :class:`pymbolic.primitives.Slice` as of its indices.
    """
    from pymbolic.primitives import Slice, Subscript
    assert isinstance(subscript, Subscript)
    return any(isinstance(index, Slice) for index in subscript.index_tuple)


def is_array_call(assignees, expression):
    """
    Returns *True* is the instruction is an array call.

    An array call is a function call applied to array type objects. If any of
    the arguments or assignees to the function is an array,
    :meth:`is_array_call` will return *True*.
    """
    from pymbolic.primitives import Call, Subscript

    from loopy.symbolic import SubArrayRef

    if not isinstance(expression, Call):
        return False

    for par in expression.parameters+assignees:
        if isinstance(par, SubArrayRef):
            return True
        elif isinstance(par, Subscript):
            if subscript_contains_slice(par):
                return True

    # did not encounter SubArrayRef/Slice, hence must be a normal call
    return False


def modify_assignee_for_array_call(assignee):
    """
    Converts the assignee subscript or variable as a SubArrayRef.
    """
    from pymbolic.primitives import Subscript, Variable

    from loopy.symbolic import SubArrayRef
    if isinstance(assignee, SubArrayRef):
        return assignee
    elif isinstance(assignee, Subscript):
        if subscript_contains_slice(assignee):
            # Slice subscripted array are treated as SubArrayRef in the kernel
            # Hence, making the behavior similar to that of `SubArrayref`
            return assignee
        else:
            return SubArrayRef((), assignee)
    elif isinstance(assignee, Variable):
        return SubArrayRef((), Subscript(assignee, 0))
    else:
        raise LoopyError("ArrayCall only takes Variable, Subscript or "
                "SubArrayRef as its inputs")


def make_assignment(assignees: tuple[Expression, ...],
                    expression: Expression,
                    temp_var_types: (
                        Sequence[LoopyType | None] | None) = None,
                    **kwargs: Any) -> Assignment | CallInstruction:

    if temp_var_types is not None:
        tv_types: Sequence[LoopyType | LoopyOptional | None] = temp_var_types
    else:
        tv_types = (LoopyOptional(),) * len(assignees)

    if len(assignees) != 1 or is_array_call(assignees, expression):
        atomicity = kwargs.pop("atomicity", ())
        if atomicity:
            raise LoopyError("atomic operations with more than one "
                    "left-hand side not supported")

        from pymbolic.primitives import Call

        from loopy.symbolic import Reduction
        if not isinstance(expression, (Call, Reduction)):
            raise LoopyError("right-hand side in multiple assignment must be "
                    "function call or reduction, got: "
                    f"'{type(expression).__name__}'")

        if not is_array_call(assignees, expression):
            return CallInstruction(
                    assignees=assignees,
                    expression=expression,
                    temp_var_types=temp_var_types,
                    **kwargs)
        else:
            # In the case of an array call, it is important to have each
            # assignee as an instance of SubArrayRef. If not given as a
            # SubArrayRef
            return CallInstruction(
                    assignees=tuple(modify_assignee_for_array_call(
                        assignee) for assignee in assignees),
                    expression=expression,
                    temp_var_types=tuple(tv_types),
                    **kwargs)
    else:
        def _is_array(expr):
            from pymbolic.primitives import Slice, Subscript

            from loopy.symbolic import SubArrayRef
            if isinstance(expr, SubArrayRef):
                return True
            if isinstance(expr, Subscript):
                return any(isinstance(idx, Slice) for idx in
                        expr.index_tuple)
            return False

        from loopy.symbolic import DependencyMapper
        if any(_is_array(dep) for dep in DependencyMapper()((assignees,
                expression))):
            raise LoopyError("Array calls only supported as instructions"
                    " with function call as RHS for now.")

        assignee, = assignees
        tv_type, = tv_types

        return Assignment(
                assignee=assignee,
                expression=expression,
                temp_var_type=tv_type,
                **kwargs)


# {{{ c instruction

class CInstruction(InstructionBase):
    """
    .. attribute:: iname_exprs

        A tuple of tuples *(name, expr)* of inames or expressions based on them
        that the instruction needs access to.

    .. attribute:: code

        The C code to be executed.

        The code should obey the following rules:

        * It should only write to temporary variables, specifically the
          temporary variables

        .. note::

            Of course, nothing in :mod:`loopy` will prevent you from doing
            'forbidden' things in your C code. If you ignore the rules and
            something breaks, you get to keep both pieces.

    .. attribute:: read_variables

        A :class:`frozenset` of variable names that :attr:`code` reads. This is
        optional and only used for figuring out dependencies.

    .. attribute:: assignees

        A sequence (typically a :class:`tuple`) of variable references (with or
        without subscript) as :data:`pymbolic.typing.Expression` instances
        that :attr:`code` writes to. This is optional and only used for
        figuring out dependencies.
    """

    fields = InstructionBase.fields | \
            set("iname_exprs code read_variables assignees".split())

    def __init__(self,
            iname_exprs, code,
            read_variables=frozenset(), assignees=(),
            id=None, happens_after=None, depends_on_is_final=None,
            groups=None, conflicts_with_groups=None,
            no_sync_with=None,
            within_inames_is_final=None, within_inames=None,
            priority=0,
            predicates=frozenset(), tags=None,
            depends_on=None):
        """
        :arg iname_exprs: Like :attr:`iname_exprs`, but instead of tuples,
            simple strings pepresenting inames are also allowed. A single
            string is also allowed, which should consists of comma-separated
            inames.
        :arg assignees: Like :attr:`assignees`, but may also be a
            semicolon-separated string of such expressions or a
            sequence of strings parseable into the desired format.
        """

        InstructionBase.__init__(self,
                id=id,
                happens_after=happens_after,
                depends_on_is_final=depends_on_is_final,
                groups=groups, conflicts_with_groups=conflicts_with_groups,
                no_sync_with=no_sync_with,
                within_inames_is_final=within_inames_is_final,
                within_inames=within_inames,
                priority=priority, predicates=predicates, tags=tags,
                depends_on=depends_on)

        # {{{ normalize iname_exprs

        if isinstance(iname_exprs, str):
            iname_exprs = [i.strip() for i in iname_exprs.split(",")]
            iname_exprs = [i for i in iname_exprs if i]

        from pymbolic import var
        new_iname_exprs = []
        for i in iname_exprs:
            if isinstance(i, str):
                new_iname_exprs.append((i, var(i)))
            else:
                new_iname_exprs.append(i)

        # }}}

        # {{{ normalize assignees

        if isinstance(assignees, str):
            assignees = [i.strip() for i in assignees.split(";")]
            assignees = [i for i in assignees if i]

        new_assignees = []
        from loopy.symbolic import parse
        for i in assignees:
            if isinstance(i, str):
                new_assignees.append(parse(i))
            else:
                new_assignees.append(i)
        # }}}

        self.iname_exprs = tuple(new_iname_exprs)
        from loopy.tools import remove_common_indentation
        self.code = remove_common_indentation(code)
        self.read_variables = read_variables
        self.assignees = tuple(new_assignees)

    # {{{ abstract interface

    def read_dependency_names(self):
        result = (
                super().read_dependency_names()
                | frozenset(self.read_variables))

        from loopy.symbolic import get_dependencies
        for _name, iname_expr in self.iname_exprs:
            result = result | get_dependencies(iname_expr)

        for subscript_deps in self.assignee_subscript_deps():
            result = result | subscript_deps

        for pred in self.predicates:
            result = result | get_dependencies(pred)

        return frozenset(result)

    def reduction_inames(self):
        return frozenset()

    def sub_array_ref_inames(self):
        return frozenset()

    def assignee_var_names(self):
        return tuple(_get_assignee_var_name(expr) for expr in self.assignees)

    def assignee_subscript_deps(self):
        return tuple(
                _get_assignee_subscript_deps(a)
                for a in self.assignees)

    def with_transformed_expressions(self, f, assignee_f=None):
        if assignee_f is None:
            assignee_f = f

        return self.copy(
                iname_exprs=[
                    (name, f(expr))
                    for name, expr in self.iname_exprs],
                assignees=[assignee_f(a) for a in self.assignees],
                predicates=frozenset(
                    f(pred) for pred in self.predicates))

    # }}}

    def __str__(self):
        first_line = "{}: {} <- CODE({}|{})".format(self.id,
                ", ".join(str(a) for a in self.assignees),
                ", ".join(str(x) for x in self.read_variables),
                ", ".join(f"{name}={expr}"
                    for name, expr in self.iname_exprs))

        options = self.get_str_options()
        if options:
            first_line += " {%s}" % (": ".join(options))

        return first_line + "\n    " + "\n    ".join(
                self.code.split("\n"))

# }}}


class _DataObliviousInstruction(InstructionBase):
    # {{{ abstract interface

    # read_dependency_names inherited

    def reduction_inames(self):
        return frozenset()

    def sub_array_ref_inames(self):
        return frozenset()

    def assignee_var_names(self):
        return frozenset()

    def assignee_subscript_deps(self):
        return frozenset()

    def with_transformed_expressions(self, f, assignee_f=None):
        return self.copy(
                predicates=frozenset(
                    f(pred) for pred in self.predicates))

    # }}}

    @property
    def assignees(self):
        return ()


# {{{ barrier instruction

class NoOpInstruction(_DataObliviousInstruction):
    """An instruction that carries out no operation. It is mainly
    useful as a way to structure dependencies between other
    instructions.

    The textual syntax in a :mod:`loopy` kernel is::

        ... nop
    """

    def __init__(self, id=None, happens_after=None, depends_on_is_final=None,
            groups=None, conflicts_with_groups=None,
            no_sync_with=None,
            within_inames_is_final=None, within_inames=None,
            priority=None,
            predicates=None, tags=None, depends_on=None):
        super().__init__(
                id=id,
                happens_after=happens_after,
                depends_on_is_final=depends_on_is_final,
                groups=groups,
                conflicts_with_groups=conflicts_with_groups,
                no_sync_with=no_sync_with,
                within_inames_is_final=within_inames_is_final,
                within_inames=within_inames,
                priority=priority,
                predicates=predicates,
                tags=tags,
                depends_on=depends_on)

    def __str__(self):
        first_line = "%s: ... nop" % self.id

        options = self.get_str_options()
        if options:
            first_line += " {%s}" % (": ".join(options))

        return first_line

# }}}


# {{{ barrier instruction

class BarrierInstruction(_DataObliviousInstruction):
    """An instruction that requires synchronization with all
    concurrent work items of :attr:`synchronization_kind`.

    .. attribute:: synchronization_kind

        A string, ``"global"`` or ``"local"``.

    .. attribute:: mem_kind

        A string, ``"global"`` or ``"local"``. Chooses which memory type to
        synchronize, for targets that require this (e.g. OpenCL)

    The textual syntax in a :mod:`loopy` kernel is::

        ... gbarrier
        ... lbarrier

    Note that the memory type :attr:`mem_kind` can be specified for local barriers::

        ... lbarrier {mem_kind=global}
    """

    fields = _DataObliviousInstruction.fields | {"synchronization_kind",
                                                     "mem_kind"}

    def __init__(self, id, happens_after=None, depends_on_is_final=None,
            groups=None, conflicts_with_groups=None,
            no_sync_with=None,
            within_inames_is_final=None, within_inames=None,
            priority=None,
            predicates=None, tags=None, synchronization_kind="global",
            mem_kind="local",
            depends_on=None):

        if predicates:
            raise LoopyError("conditional barriers are not supported")

        super().__init__(
                id=id,
                happens_after=happens_after,
                depends_on_is_final=depends_on_is_final,
                groups=groups,
                conflicts_with_groups=conflicts_with_groups,
                no_sync_with=no_sync_with,
                within_inames_is_final=within_inames_is_final,
                within_inames=within_inames,
                priority=priority,
                predicates=predicates,
                tags=tags,
                depends_on=depends_on)

        self.synchronization_kind = synchronization_kind
        self.mem_kind = mem_kind

    def __str__(self):
        first_line = \
                "{}: ... {}barrier".format(self.id, self.synchronization_kind[0])

        options = self.get_str_options()
        if self.synchronization_kind == "local":
            # add the memory kind
            options += [f"mem_kind={self.mem_kind}"]
        if options:
            first_line += " {%s}" % (": ".join(options))

        return first_line

# }}}


# {{{ key getters

def _get_insn_eq_key(insn):
    return insn._key_builder.key()


def _get_insn_hash_key(insn):
    return insn._key_builder.hash_key()

# }}}


# {{{ _check_and_fix_temp_var_type

def _check_and_fix_temp_var_type(temp_var_type, stacklevel=2):
    """Check temp_var_type for deprecated usage, and convert to the right value.
    """

    import loopy as lp

    if temp_var_type is None:
        warn("temp_var_type should be Optional() if no temporary, not None. "
             "This usage will be disallowed soon.",
             DeprecationWarning, stacklevel=1 + stacklevel)
        temp_var_type = lp.Optional()

    elif temp_var_type is lp.auto:
        warn("temp_var_type should be Optional(None) if "
             "unspecified, not auto. This usage will be disallowed soon.",
             DeprecationWarning, stacklevel=1 + stacklevel)
        temp_var_type = lp.Optional(None)

    elif not isinstance(temp_var_type, lp.Optional):
        warn("temp_var_type should be an instance of Optional. "
             "Other values for temp_var_type will be disallowed soon.",
             DeprecationWarning, stacklevel=1 + stacklevel)
        temp_var_type = lp.Optional(temp_var_type)

    return temp_var_type

# }}}


def get_insn_domain(insn, kernel):
    """
    Returns an instance of :class:`islpy.Set` for the *insn*'s domain.

    .. note::

        Does not take into account additional hints available through
        :attr:`loopy.LoopKernel.assumptions`.
    """
    domain = kernel.get_inames_domain(insn.within_inames)

    # {{{ add read-only ValueArgs to domain

    from loopy.kernel.data import ValueArg

    valueargs_to_add = ({arg.name for arg in kernel.args
                         if isinstance(arg, ValueArg)
                         and arg.name not in kernel.get_written_variables()}
                        - set(domain.get_var_names(isl.dim_type.param)))

    # only consider valueargs relevant to *insn*
    valueargs_to_add = valueargs_to_add & insn.read_dependency_names()

    for arg_to_add in valueargs_to_add:
        idim = domain.dim(isl.dim_type.param)
        domain = domain.add_dims(isl.dim_type.param, 1)
        domain = domain.set_dim_name(isl.dim_type.param, idim, arg_to_add)

    # }}}

    # {{{ enforce restriction from predicates

    insn_preds_set = isl.BasicSet.universe(domain.space)

    for predicate in insn.predicates:
        from loopy.symbolic import condition_to_set
        predicate_as_isl_set = condition_to_set(domain.space, predicate)
        if predicate_as_isl_set is not None:
            insn_preds_set = insn_preds_set & predicate_as_isl_set

    # }}}

    return domain & insn_preds_set


# vim: foldmethod=marker



"""Data used by the kernel object."""

from __future__ import annotations


__copyright__ = "Copyright (C) 2012 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


from dataclasses import dataclass, replace
from enum import IntEnum
from sys import intern
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Sequence,
    Tuple,
    Union,
    cast,
)

import numpy  # FIXME: imported as numpy to allow sphinx to resolve things
import numpy as np

from pytools import ImmutableRecord
from pytools.tag import Tag, Taggable, TagT, UniqueTag as UniqueTagBase

from loopy.diagnostic import LoopyError
from loopy.kernel.array import ArrayBase, ArrayDimImplementationTag
from loopy.kernel.instruction import (  # noqa
    Assignment,
    AtomicInit,
    AtomicUpdate,
    CallInstruction,
    CInstruction,
    InstructionBase,
    MemoryOrdering,
    MemoryScope,
    MultiAssignmentBase,
    VarAtomicity,
    make_assignment,
)
from loopy.typing import Expression, ShapeType, auto


if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping

    from pymbolic import ArithmeticExpression, Variable

    from loopy.types import LoopyType, ToLoopyTypeConvertible


__doc__ = """
.. autofunction:: filter_iname_tags_by_type

.. autoclass:: InameImplementationTag

.. autoclass:: ConcurrentTag

.. autoclass:: UniqueInameTag

.. autoclass:: AxisTag

.. autoclass:: LocalInameTag

.. autoclass:: GroupInameTag

.. autoclass:: VectorizeTag

.. autoclass:: UnrollTag

.. autoclass:: Iname

References
^^^^^^^^^^

.. class:: ToLoopyTypeConvertible

    See :class:`loopy.ToLoopyTypeConvertible`.

.. class:: TagT

    A type variable with a lower bound of :class:`pytools.tag.Tag`.
"""

# This docstring is included in ref_internals. Do not include parts of the public
# interface, e.g. TemporaryVariable, KernelArgument, ArrayArg.


# {{{ utilities

def _names_from_expr(expr: Expression | str | None) -> frozenset[str]:
    from numbers import Number

    from loopy.symbolic import DependencyMapper
    dep_mapper: DependencyMapper[[]] = DependencyMapper()

    from pymbolic.primitives import ExpressionNode
    if isinstance(expr, str):
        return frozenset({expr})
    elif isinstance(expr, ExpressionNode):
        return frozenset(cast("Variable", v).name for v in dep_mapper(expr))
    elif expr is None:
        return frozenset()
    elif isinstance(expr, Number):
        return frozenset()
    else:
        raise ValueError(f"unexpected value of expression-like object: '{expr}'")


def _names_from_dim_tags(
        dim_tags: Sequence[ArrayDimImplementationTag] | None) -> frozenset[str]:
    from loopy.kernel.array import FixedStrideArrayDimTag
    if dim_tags is not None:
        return frozenset({
            name
            for dim_tag in dim_tags
            if isinstance(dim_tag, FixedStrideArrayDimTag)
            for name in _names_from_expr(dim_tag.stride)
            })
    else:
        return frozenset()

# }}}


# {{{ iname tags

def filter_iname_tags_by_type(
            tags: Iterable[Tag],
            tag_type: type[TagT] | tuple[type[TagT], ...],
            max_num: int | None = None,
            min_num: int | None = None,
        ) -> set[TagT]:
    """Return a subset of *tags* that matches type *tag_type*. Raises exception
    if the number of tags found were greater than *max_num* or less than
    *min_num*.

    :arg tags: An iterable of tags.
    :arg tag_type: a subclass of :class:`loopy.kernel.data.InameImplementationTag`.
    :arg max_num: the maximum number of tags expected to be found.
    :arg min_num: the minimum number of tags expected to be found.
    """

    result: set[TagT] = cast(
        "set[TagT]",
        {tag for tag in tags if isinstance(tag, tag_type)})

    def strify_tag_type():
        if isinstance(tag_type, tuple):
            return ", ".join(t.__name__ for t in tag_type)
        else:
            return tag_type.__name__

    if max_num is not None:
        if len(result) > max_num:
            raise LoopyError("cannot have more than {} tags "
                    "of type(s): {}".format(max_num, strify_tag_type()))
    if min_num is not None:
        if len(result) < min_num:
            raise LoopyError("must have more than {} tags "
                    "of type(s): {}".format(max_num, strify_tag_type()))

    return result


class InameImplementationTag(ImmutableRecord, UniqueTagBase):
    __slots__: ClassVar[tuple[str, ...]] = ()

    def __hash__(self):
        return hash(self.key)

    def __lt__(self, other):
        return self.__hash__() < other.__hash__()

    def update_persistent_hash(self, key_hash, key_builder):
        """Custom hash computation function for use with
        :class:`pytools.persistent_dict.PersistentDict`.
        """

        return key_builder.rec(key_hash, self.key)

    @property
    def key(self):
        """Return a hashable, comparable value that is used to ensure
        per-instruction uniqueness of this unique iname tag.

        Also used for persistent hash construction.
        """
        return type(self).__name__


class ConcurrentTag(InameImplementationTag):
    pass


class HardwareConcurrentTag(ConcurrentTag):
    pass


class UniqueInameTag(InameImplementationTag):
    pass


class AxisTag(UniqueInameTag):
    __slots__ = ["axis"]

    def __init__(self, axis):
        ImmutableRecord.__init__(self,
                axis=axis)

    @property
    def key(self):
        return (type(self).__name__, self.axis)

    def __str__(self):
        return "%s.%d" % (
                self.print_name, self.axis)


class GroupInameTag(HardwareConcurrentTag, AxisTag):
    print_name = "g"


class LocalInameTagBase(HardwareConcurrentTag):
    pass


class LocalInameTag(LocalInameTagBase, AxisTag):
    print_name = "l"


class AutoLocalInameTagBase(LocalInameTagBase):
    @property
    def key(self):
        return type(self).__name__


class AutoFitLocalInameTag(AutoLocalInameTagBase):
    def __str__(self):
        return "l.auto"


# {{{ ilp-like

class IlpBaseTag(ConcurrentTag):
    pass


class UnrolledIlpTag(IlpBaseTag):
    def __str__(self):
        return "ilp.unr"


class LoopedIlpTag(IlpBaseTag):
    def __str__(self):
        return "ilp.seq"

# }}}


class VectorizeTag(UniqueInameTag, HardwareConcurrentTag):
    def __str__(self):
        return "vec"


class UnrollTag(InameImplementationTag):
    def __str__(self):
        return "unr"


class UnrollHintTag(InameImplementationTag):
    __slots__ = ["value"]

    def __init__(self, value=None):
        ImmutableRecord.__init__(self,
                value=value)

    @property
    def key(self):
        return (type(self).__name__, self.value)

    def __str__(self):
        if self.value:
            return f"unr_hint.{self.value}"
        else:
            return "unr_hint"


class ForceSequentialTag(InameImplementationTag):
    def __str__(self):
        return "forceseq"


class InOrderSequentialSequentialTag(InameImplementationTag):
    def __str__(self):
        return "ord"


ToInameTagConvertible = Union[str, Tag, None]


def parse_tag(tag: ToInameTagConvertible) -> Tag | None:
    if tag is None:
        return tag

    if isinstance(tag, Tag):
        return tag

    if not isinstance(tag, str):
        raise ValueError("cannot parse tag: %s" % tag)

    if tag == "for":
        return None
    elif tag == "ord":
        return InOrderSequentialSequentialTag()
    elif tag in ["unr"]:
        return UnrollTag()
    elif tag in ["vec"]:
        return VectorizeTag()
    elif tag in ["ilp", "ilp.unr"]:
        return UnrolledIlpTag()
    elif tag == "ilp.seq":
        return LoopedIlpTag()
    elif tag == "unr_hint":
        return UnrollHintTag()
    elif tag.startswith("unr_hint."):
        offset = len("unr_hint.")
        return UnrollHintTag(int(tag[offset:]))
    elif tag.startswith("g."):
        return GroupInameTag(int(tag[2:]))
    elif tag.startswith("l."):
        axis = tag[2:]
        if axis == "auto":
            return AutoFitLocalInameTag()
        else:
            return LocalInameTag(int(axis))
    else:
        raise ValueError("cannot parse tag: %s" % tag)

# }}}


# {{{ memory address space

class AddressSpace(IntEnum):
    """Storage location of a variable.

    .. attribute:: PRIVATE
    .. attribute:: LOCAL
    .. attribute:: GLOBAL
    """

    # These must occur in ascending order of 'globality' so that
    # max(scope) does the right thing.

    PRIVATE = 0
    LOCAL = 1
    GLOBAL = 2

    @classmethod
    def stringify(cls, val: AddressSpace | type[auto]) -> str:
        if val == cls.PRIVATE:
            return "private"
        elif val == cls.LOCAL:
            return "local"
        elif val == cls.GLOBAL:
            return "global"
        elif val is auto:
            return "<auto>"
        else:
            raise ValueError("unexpected value of AddressSpace")

# }}}


# {{{ arguments

class KernelArgument(ImmutableRecord):
    """Base class for all argument types.

    .. attribute:: name
    .. attribute:: dtype
    .. attribute:: is_output
    .. attribute:: is_input

    .. automethod:: supporting_names
    """
    name: str
    dtype: LoopyType | None
    is_output: bool
    is_input: bool

    def __init__(self, **kwargs):
        kwargs["name"] = intern(kwargs.pop("name"))

        dtype = kwargs.pop("dtype", None)

        for_atomic = kwargs.pop("for_atomic", False)

        from loopy.types import to_loopy_type
        dtype = to_loopy_type(
                dtype, allow_auto=True, allow_none=True, for_atomic=for_atomic)

        import loopy as lp
        if dtype is lp.auto:
            raise TypeError("dtype may not be lp.auto")

        kwargs["dtype"] = dtype
        kwargs["is_output"] = kwargs.pop("is_output", None)
        kwargs["is_input"] = kwargs.pop("is_input", None)

        ImmutableRecord.__init__(self, **kwargs)

    def supporting_names(self) -> frozenset[str]:
        """'Supporting' names are those that are likely to be required to be
        present for any use of the argument.
        """

        return frozenset()


@dataclass(frozen=True)
class _ArraySeparationInfo:
    """Not user-facing. If an array has been split because an axis
    is tagged with :class:`~loopy.kernel.data.SeparateArrayArrayDimTag`,
    this records the names of the actually present sub-arrays that
    should be used to realize this array.
    """
    sep_axis_indices_set: frozenset[int]
    subarray_names: Mapping[tuple[int, ...], str]


class ArrayArg(ArrayBase, KernelArgument):
    __doc__ = cast("str", ArrayBase.__doc__) + (
        """
        .. attribute:: address_space

            An attribute of :class:`AddressSpace` defining the address
            space in which the array resides.

        .. attribute:: is_output

            An instance of :class:`bool`. If set to *True*, the array is used to
            return information to the caller. If set to *False*, the callee does not
            write to the array during a call.

        .. attribute:: is_input

            An instance of :class:`bool`. If set to *True*, expected to be provided
            by the caller. If *False*, the callee does not depend on the array
            at kernel entry.
        """)

    address_space: AddressSpace

    # _separation_info is not user-facing and hence not documented.
    _separation_info: _ArraySeparationInfo | None

    allowed_extra_kwargs = (
            "address_space",
            "is_output",
            "is_input",
            "tags",
            "_separation_info")

    def __init__(self, *args, **kwargs):
        if "address_space" not in kwargs:
            raise TypeError("'address_space' must be specified")

        kwargs["is_output"] = kwargs.pop("is_output", None)
        kwargs["is_input"] = kwargs.pop("is_input", None)
        kwargs["_separation_info"] = kwargs.pop("_separation_info", None)

        super().__init__(*args, **kwargs)

    min_target_axes = 0
    max_target_axes = 1

    def __str__(self):
        # Don't mention the type of array arg if shape is known
        # FIXME: Why?
        include_typename = self.shape in (None, auto)

        aspace_str = AddressSpace.stringify(self.address_space)

        inout = []
        if self.is_input:
            inout.append("in")
        if self.is_output:
            inout.append("out")
        if not (self.is_input or self.is_output):
            inout.append("neither_in_nor_out?")

        return (
                self.stringify(include_typename=include_typename)
                + " " + "/".join(inout)
                + " aspace: %s" % aspace_str)

    def update_persistent_hash(self, key_hash, key_builder):
        """Custom hash computation function for use with
        :class:`pytools.persistent_dict.PersistentDict`.
        """
        super().update_persistent_hash(key_hash, key_builder)
        key_builder.rec(key_hash, self.address_space)
        key_builder.rec(key_hash, self.is_output)
        key_builder.rec(key_hash, self.is_input)
        key_builder.rec(key_hash, self._separation_info)

    def supporting_names(self) -> frozenset[str]:
        # Do not consider separation info here: The subarrays don't support, they
        # replace this array.
        return (
                _names_from_expr(self.offset)
                | _names_from_dim_tags(self.dim_tags)
                )


# Making this a function prevents incorrect use in isinstance.
# Note: This is *not* deprecated, as it is super-common and
# incrementally more convenient to use than ArrayArg directly.
def GlobalArg(*args, **kwargs) -> ArrayArg:  # noqa: N802
    address_space = kwargs.pop("address_space", None)
    if address_space is not None:
        raise TypeError("may not pass 'address_space' to GlobalArg")
    kwargs["address_space"] = AddressSpace.GLOBAL

    return ArrayArg(*args, **kwargs)


class ConstantArg(ArrayBase, KernelArgument):
    __doc__ = ArrayBase.__doc__

    def __init__(self, *args, **kwargs):
        if kwargs.pop("address_space", AddressSpace.GLOBAL) != AddressSpace.GLOBAL:
            raise LoopyError("'address_space' for ConstantArg must be GLOBAL.")
        super().__init__(*args, **kwargs)

    # Constant Arg cannot be an output
    is_output = False
    is_input = True
    address_space = AddressSpace.GLOBAL

    min_target_axes = 0
    max_target_axes = 1


class ImageArg(ArrayBase, KernelArgument):
    __doc__ = ArrayBase.__doc__

    def __init__(self, *args, **kwargs):
        if kwargs.pop("address_space", AddressSpace.GLOBAL) != AddressSpace.GLOBAL:
            raise LoopyError("'address_space' for ImageArg must be GLOBAL.")
        super().__init__(*args, **kwargs)

    min_target_axes = 1
    max_target_axes = 3

    # ImageArg cannot be an output (for now)
    is_output = False
    is_input = True
    address_space = AddressSpace.GLOBAL

    @property
    def dimensions(self):
        assert self.dim_tags is not None
        return len(self.dim_tags)

    def get_arg_decl(self, ast_builder, name_suffix, shape, dtype, is_written):
        return ast_builder.get_image_arg_decl(self.name + name_suffix, shape,
                self.num_target_axes(), dtype, is_written)

    def supporting_names(self) -> frozenset[str]:
        return (
                _names_from_expr(self.offset)
                | _names_from_dim_tags(self.dim_tags)
                )


class ValueArg(KernelArgument, Taggable):
    def __init__(self,
                name: str,
                dtype: ToLoopyTypeConvertible | None = None,
                approximately: int = 1000,
                is_output: bool = False,
                is_input: bool = True,
                tags: frozenset[Tag] | None = None,
             ) -> None:
        """
        :arg tags: A an instance of or Iterable of instances of
            :class:`pytools.tag.Tag` intended for consumption by an
            application.
        """

        if tags is None:
            tags = frozenset()

        KernelArgument.__init__(self, name=name,
                dtype=dtype,
                approximately=approximately,
                is_output=is_output,
                is_input=is_input,
                tags=tags)

    def __str__(self):
        import loopy as lp
        assert self.dtype is not lp.auto

        if self.dtype is None:
            type_str = "<auto/runtime>"
        else:
            type_str = str(self.dtype)

        return f"{self.name}: ValueArg, type: {type_str}"

    def __repr__(self):
        return "<%s>" % self.__str__()

    def update_persistent_hash(self, key_hash, key_builder):
        """Custom hash computation function for use with
        :class:`pytools.persistent_dict.PersistentDict`.
        """

        key_builder.rec(key_hash, self.name)
        key_builder.rec(key_hash, self.dtype)

    def get_arg_decl(self, ast_builder):
        return ast_builder.get_value_arg_decl(self.name, (),
                self.dtype, False)

# }}}


# {{{ temporary variable

class TemporaryVariable(ArrayBase):
    __doc__ = cast("str", ArrayBase.__doc__) + """
    .. autoattribute:: storage_shape
    .. autoattribute:: base_indices
    .. autoattribute:: address_space
    .. autoattribute:: base_storage
    .. autoattribute:: initializer
    .. autoattribute:: read_only
    .. autoattribute:: _base_storage_access_may_be_aliasing
    """

    storage_shape: ShapeType | None
    base_indices: tuple[Expression, ...] | None
    address_space: AddressSpace | type[auto]
    base_storage: str | None
    """The name of a storage array that is to be used to actually
    hold the data in this temporary, or *None*. If not *None* or the name
    of an existing variable, a variable of this name and appropriate size
    will be created.
    """

    initializer: numpy.ndarray | None
    """*None* or a :class:`numpy.ndarray` of data to be used to initialize the
    array.
    """

    read_only: bool
    """A :class:`bool` indicating whether the variable may be written during
    its lifetime. If *True*, *initializer* must be given.
    """

    _base_storage_access_may_be_aliasing: bool
    """Whether the temporary is used to alias the underlying base storage.
    Defaults to *False*. If *False*, C-based code generators will declare
    the temporary as a ``restrict`` const pointer to the base storage
    memory location. If *True*, the restrict part is omitted on this
    declaration.
    """

    min_target_axes: ClassVar[int] = 0
    max_target_axes: ClassVar[int] = 1

    allowed_extra_kwargs = (
            "storage_shape",
            "base_indices",
            "address_space",
            "base_storage",
            "initializer",
            "read_only",
            "_base_storage_access_may_be_aliasing",
            )

    def __init__(
                self,
                name: str,
                dtype: ToLoopyTypeConvertible = None,
                shape: ShapeType | type[auto] | None = auto,
                address_space: AddressSpace | type[auto] | None = None,
                dim_tags: Sequence[ArrayDimImplementationTag] | None = None,
                offset: Expression | str | None = 0,
                dim_names: tuple[str, ...] | None = None,
                strides: tuple[Expression, ...] | None = None,
                order: str | None = None,

                base_indices: tuple[Expression, ...] | None = None,
                storage_shape: ShapeType | None = None,

                base_storage: str | None = None,
                initializer: np.ndarray | None = None,
                read_only: bool = False,

                _base_storage_access_may_be_aliasing: bool = False,
                **kwargs: Any
            ) -> None:
        """
        :arg dtype: :class:`loopy.auto` or a :class:`numpy.dtype`
        :arg shape: :class:`loopy.auto` or a shape tuple
        :arg base_indices: :class:`loopy.auto` or a tuple of base indices
        """

        if address_space is None:
            address_space = auto

        if initializer is None:
            pass
        elif isinstance(initializer, np.ndarray):
            if offset != 0:
                raise LoopyError(
                        "temporary variable '%s': "
                        "offset must be 0 if initializer specified"
                        % name)

            from loopy.types import NumpyType, to_loopy_type
            if dtype is auto or dtype is None:
                dtype = NumpyType(initializer.dtype)
            elif to_loopy_type(dtype) != to_loopy_type(initializer.dtype):
                raise LoopyError(
                        "temporary variable '%s': "
                        "dtype of initializer does not match "
                        "dtype of array."
                        % name)

            if shape is auto:
                shape = initializer.shape
            else:
                if shape != initializer.shape:
                    raise LoopyError("Shape of '{}' does not match that of the"
                            " initializer.".format(name))
        else:
            raise LoopyError(
                    "temporary variable '%s': "
                    "initializer must be None or a numpy array"
                    % name)

        if order is None:
            order = "C"

        if shape is not None:
            from loopy.kernel.array import _parse_shape_or_strides
            shape = _parse_shape_or_strides(shape)

        if base_indices is None and shape is not auto and shape is not None:
            assert isinstance(shape, tuple)
            base_indices = (0,) * len(shape)

        if not read_only and initializer is not None:
            raise LoopyError(
                    "temporary variable '%s': "
                    "read-write variables with initializer "
                    "are not currently supported "
                    "(did you mean to set read_only=True?)"
                    % name)

        if base_storage is not None and initializer is not None:
            raise LoopyError(
                    "temporary variable '%s': "
                    "base_storage and initializer are "
                    "mutually exclusive"
                    % name)

        if base_storage is None and _base_storage_access_may_be_aliasing:
            raise LoopyError(
                    "temporary variable '%s': "
                    "_base_storage_access_may_be_aliasing option, but no "
                    "base_storage given!"
                    % name)

        ArrayBase.__init__(self, name=intern(name),
                dtype=dtype, shape=shape, strides=strides,
                dim_tags=dim_tags, offset=offset, dim_names=dim_names,
                order=order,
                base_indices=base_indices,
                address_space=address_space,
                storage_shape=storage_shape,
                base_storage=base_storage,
                initializer=initializer,
                read_only=read_only,
                _base_storage_access_may_be_aliasing=(
                    _base_storage_access_may_be_aliasing),
                **kwargs)

    def copy(self, **kwargs: Any) -> TemporaryVariable:
        address_space = kwargs.pop("address_space", None)

        if address_space is not None:
            kwargs["address_space"] = address_space

        return super().copy(**kwargs)

    @property
    def nbytes(self) -> Expression:
        if self.storage_shape is not None:
            shape = self.storage_shape
        else:
            if self.shape is None:
                raise ValueError("shape is None")
            if self.shape is auto:
                raise ValueError("shape is auto")
            shape = cast("Tuple[ArithmeticExpression]", self.shape)

        if self.dtype is None:
            raise ValueError("data type is indeterminate")

        from pytools import product
        return product(si for si in shape)*self.dtype.itemsize

    def __str__(self) -> str:
        if self.address_space is auto:
            aspace_str = "auto"
        else:
            aspace_str = AddressSpace.stringify(self.address_space)

        if self.base_storage is None:
            bs_str = ""
        else:
            bs_str = " base_storage: "+str(self.base_storage)

        return (
                self.stringify(include_typename=False)
                + f" aspace: {aspace_str}{bs_str}")

    def __eq__(self, other):
        return (
                super().__eq__(other)
                and self.storage_shape == other.storage_shape
                and self.base_indices == other.base_indices
                and self.address_space == other.address_space
                and self.base_storage == other.base_storage
                and (
                    (self.initializer is None and other.initializer is None)
                    or np.array_equal(self.initializer, other.initializer))
                and self.read_only == other.read_only
                and (self._base_storage_access_may_be_aliasing
                    == other._base_storage_access_may_be_aliasing)
                )

    def update_persistent_hash(self, key_hash, key_builder):
        """Custom hash computation function for use with
        :class:`pytools.persistent_dict.PersistentDict`.
        """

        super().update_persistent_hash(key_hash, key_builder)
        key_builder.rec(key_hash, self.storage_shape)
        key_builder.rec(key_hash, self.base_indices)
        key_builder.rec(key_hash, self.address_space)
        key_builder.rec(key_hash, self.base_storage)

        initializer = self.initializer
        if initializer is not None:
            initializer = (initializer.tolist(), initializer.dtype)
        key_builder.rec(key_hash, initializer)

        key_builder.rec(key_hash, self.read_only)
        key_builder.rec(key_hash, self._base_storage_access_may_be_aliasing)

    def supporting_names(self) -> frozenset[str]:
        return (
                _names_from_expr(self.offset)
                | _names_from_dim_tags(self.dim_tags)
                | (
                    frozenset({self.base_storage})
                    if self.base_storage else frozenset())
                )

# }}}


# {{{ substitution rule

@dataclass(frozen=True)
class SubstitutionRule:
    """
    .. autoattribute:: name
    .. autoattribute:: arguments
    .. autoattribute:: expression
    """

    name: str
    arguments: Sequence[str]
    expression: Expression

    def copy(self, **kwargs: Any) -> SubstitutionRule:
        return replace(self, **kwargs)

    def update_persistent_hash(self, key_hash, key_builder):
        key_builder.rec(key_hash, self.name)
        key_builder.rec(key_hash, self.arguments)
        key_builder.rec(key_hash, self.expression)


# }}}


# {{{ function call mangling

class CallMangleInfo(ImmutableRecord):
    """
    .. attribute:: target_name

        A string. The name of the function to be called in the
        generated target code.

    .. attribute:: result_dtypes

        A tuple of :class:`loopy.types.LoopyType` instances indicating what
        types of values the function returns.

    .. attribute:: arg_dtypes

        A tuple of :class:`loopy.types.LoopyType` instances indicating what
        types of arguments the function actually receives.
    """

    def __init__(self, target_name, result_dtypes, arg_dtypes):
        assert isinstance(result_dtypes, tuple)

        super().__init__(
                target_name=target_name,
                result_dtypes=result_dtypes,
                arg_dtypes=arg_dtypes)

# }}}


# {{{ Iname class

@dataclass(frozen=True)
class Iname(Taggable):
    """
    Records an iname in a :class:`~loopy.LoopKernel`. See :ref:`domain-tree` for
    semantics of *inames* in :mod:`loopy`.

    This class records the metadata attached to an iname as instances of
    :class:pytools.tag.Tag`. A tag maybe a builtin tag like
    :class:`loopy.kernel.data.InameImplementationTag` or a user-defined custom
    tag. Custom tags may be attached to inames to be used in targeting later
    during transformations.

    .. attribute:: name

        An instance of :class:`str`, denoting the iname's name.

    .. attribute:: tags

        An instance of :class:`frozenset` of :class:`pytools.tag.Tag`.
    """
    name: str
    tags: frozenset[Tag]

    def copy(self, **kwargs: Any) -> Iname:
        return replace(self, **kwargs)

    def _with_new_tags(self, tags):
        return self.copy(tags=tags)

# }}}


# vim: foldmethod=marker



from __future__ import annotations


__copyright__ = "Copyright (C) 2012 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


from typing import TYPE_CHECKING

import numpy as np

import islpy as isl
import pymbolic.primitives as p
from pymbolic import var
from pymbolic.mapper import IdentityMapper, Mapper
from pymbolic.mapper.stringifier import (
    PREC_BITWISE_AND,
    PREC_BITWISE_OR,
    PREC_BITWISE_XOR,
    PREC_CALL,
    PREC_LOGICAL_AND,
    PREC_LOGICAL_OR,
    PREC_NONE,
    PREC_PRODUCT,
    PREC_SHIFT,
    PREC_UNARY,
)

from loopy.diagnostic import LoopyError
from loopy.expression import dtype_to_type_context
from loopy.target.c import CExpression
from loopy.type_inference import TypeInferenceMapper, TypeReader
from loopy.types import LoopyType
from loopy.typing import Expression, is_integer


if TYPE_CHECKING:
    from loopy.codegen import CodeGenerationState
    from loopy.symbolic import TypeCast


__doc__ = """
.. currentmodule:: loopy.target.c.codegen.expression

.. autoclass:: ExpressionToCExpressionMapper
"""


# {{{ Loopy expression to C expression mapper

class ExpressionToCExpressionMapper(IdentityMapper):
    """
    Mapper that converts a loopy-semantic expression to a C-semantic expression
    with typecasts, appropriate arithmetic semantic mapping, etc.


    .. note::

        - All mapper methods take in an extra argument called *type_context*.
          The purpose of *type_context* is to inform the method about the
          expected type for untyped expressions such as python scalars. The
          type of the expressions takes precedence over *type_context*.
    """
    def __init__(self,
                 codegen_state: CodeGenerationState,
                 fortran_abi: bool = False,
                 type_inf_mapper: TypeInferenceMapper | None = None
             ) -> None:
        self.kernel = codegen_state.kernel
        self.codegen_state = codegen_state

        if type_inf_mapper is None:
            type_inf_mapper = TypeReader(self.kernel,
                    self.codegen_state.callables_table)
        self.type_inf_mapper = type_inf_mapper

        self.allow_complex = codegen_state.allow_complex

        self.fortran_abi = fortran_abi

    # {{{ helpers

    def with_assignments(self, names_to_vars):
        type_inf_mapper = self.type_inf_mapper.with_assignments(names_to_vars)
        return type(self)(self.codegen_state, self.fortran_abi, type_inf_mapper)

    def infer_type(self, expr: Expression) -> LoopyType:
        result = self.type_inf_mapper(expr)
        assert isinstance(result, LoopyType)

        self.codegen_state.seen_dtypes.add(result)
        return result

    def find_array(self, expr):
        if expr.aggregate.name in self.kernel.arg_dict:
            ary = self.kernel.arg_dict[expr.aggregate.name]
        elif expr.aggregate.name in self.kernel.temporary_variables:
            ary = self.kernel.temporary_variables[expr.aggregate.name]
        else:
            raise RuntimeError("nothing known about subscripted variable '%s'"
                    % expr.aggregate.name)

        from loopy.kernel.array import ArrayBase
        if not isinstance(ary, ArrayBase):
            raise RuntimeError("subscripted variable '%s' is not an array"
                    % expr.aggregate.name)

        return ary

    def wrap_in_typecast(self, actual_type: LoopyType, needed_type: LoopyType, s):
        if actual_type != needed_type:
            registry = self.codegen_state.ast_builder.target.get_dtype_registry()
            cast = var("(%s) " % registry.dtype_to_ctype(needed_type))
            return cast(s)

        return s

    def rec(self, expr, type_context=None, needed_type: LoopyType | None = None):  # type: ignore[override]
        result = super().rec(expr, type_context)

        if needed_type is None:
            return result
        else:
            return self.wrap_in_typecast(
                    self.infer_type(expr), needed_type,
                    result)

    def __call__(self, expr, prec=None, type_context=None, needed_dtype=None):
        if prec is None:
            prec = PREC_NONE

        assert prec == PREC_NONE
        return CExpression(
                self.codegen_state.ast_builder.get_c_expression_to_code_mapper(),
                self.rec(expr, type_context, needed_dtype))

    # }}}

    def map_variable(self, expr, type_context):
        from loopy.kernel.data import AddressSpace, ValueArg

        def postproc(x):
            return x

        if expr.name in self.codegen_state.var_subst_map:
            if self.kernel.options.annotate_inames:
                return var(
                        "/* {} */ {}".format(
                            expr.name,
                            self.rec(self.codegen_state.var_subst_map[expr.name],
                                type_context)))
            else:
                return self.rec(self.codegen_state.var_subst_map[expr.name],
                    type_context)
        elif expr.name in self.kernel.arg_dict:
            arg = self.kernel.arg_dict[expr.name]
            from loopy.kernel.array import ArrayBase
            if isinstance(arg, ArrayBase):
                if arg.shape == ():
                    if arg.offset:

                        from loopy.kernel.array import _apply_offset
                        from loopy.symbolic import simplify_using_aff

                        subscript = _apply_offset(0, arg)
                        result = self.make_subscript(
                                arg,
                                var(expr.name),
                                simplify_using_aff(
                                    self.kernel, self.rec(subscript, "i")))
                        return result
                    else:
                        return var(expr.name)[0]
                else:
                    raise RuntimeError("unsubscripted reference to array '%s'"
                            % expr.name)

            if isinstance(arg, ValueArg) and self.fortran_abi:
                postproc = lambda x: x[0]  # noqa
        elif expr.name in self.kernel.temporary_variables:
            temporary = self.kernel.temporary_variables[expr.name]
            if (temporary.base_storage
                    or temporary.address_space == AddressSpace.GLOBAL):
                postproc = lambda x: x[0]  # noqa

        result = self.kernel.mangle_symbol(self.codegen_state.ast_builder, expr.name)
        if result is not None:
            _, c_name = result
            return postproc(var(c_name))

        return postproc(var(expr.name))

    def map_tagged_variable(self, expr, type_context):
        return var(expr.name)

    def map_sub_array_ref(self, expr, type_context):
        from loopy.symbolic import get_start_subscript_from_sar
        return var("&")(self.rec(get_start_subscript_from_sar(expr, self.kernel),
            type_context))

    def map_subscript(self, expr, type_context):
        def base_impl(expr, type_context):
            return self.rec(expr.aggregate, type_context)[self.rec(expr.index, "i")]

        def make_var(name):
            from loopy import TaggedVariable
            if isinstance(expr.aggregate, TaggedVariable):
                return TaggedVariable(name, expr.aggregate.tags)
            else:
                return var(name)

        from pymbolic.primitives import Variable
        if not isinstance(expr.aggregate, Variable):
            return base_impl(expr, type_context)

        ary = self.find_array(expr)

        from pymbolic import evaluate

        from loopy.kernel.array import get_access_info
        from loopy.symbolic import simplify_using_aff
        index_tuple = tuple(
                simplify_using_aff(self.kernel, idx) for idx in expr.index_tuple)

        access_info = get_access_info(self.kernel, ary, index_tuple,
                lambda expr: evaluate(expr, self.codegen_state.var_subst_map),
                self.codegen_state.vectorization_info)

        from loopy.kernel.data import ArrayArg, ConstantArg, ImageArg, TemporaryVariable

        if isinstance(ary, ImageArg):
            extra_axes = 0

            num_target_axes = ary.num_target_axes()
            if num_target_axes in [1, 2]:
                idx_vec_type = "float2"
                extra_axes = 2-num_target_axes
            elif num_target_axes == 3:
                idx_vec_type = "float4"
                extra_axes = 4-num_target_axes
            else:
                raise LoopyError("unsupported number (%d) of target axes in image"
                        % num_target_axes)

            idx_tuple = expr.index_tuple[::-1] + (0,) * extra_axes

            base_access = var("read_imagef")(
                    var(ary.name),
                    var("loopy_sampler"),
                    var("(%s)" % idx_vec_type)(*self.rec(idx_tuple, "i")))

            if ary.dtype.numpy_dtype == np.float32:
                return base_access.attr("x")
            if self.kernel.target.is_vector_dtype(ary.dtype):
                return base_access
            elif ary.dtype.numpy_dtype == np.float64:
                return var("as_double")(base_access.attr("xy"))
            else:
                raise NotImplementedError(
                        "non-floating-point images not supported for now")

        elif isinstance(ary, (ArrayArg, TemporaryVariable, ConstantArg)):
            if len(access_info.subscripts) == 0:
                if (
                        isinstance(ary, (ConstantArg, ArrayArg)) or
                        (isinstance(ary, TemporaryVariable) and ary.base_storage)):
                    # unsubscripted global args are pointers
                    result = self.make_subscript(
                            ary,
                            make_var(access_info.array_name),
                            (0,))

                else:
                    # unsubscripted temp vars are scalars
                    # (unless they use base_storage)
                    result = make_var(access_info.array_name)

            else:
                subscript, = access_info.subscripts
                result = self.make_subscript(
                        ary,
                        make_var(access_info.array_name),
                        simplify_using_aff(
                            self.kernel, self.rec(subscript, "i")))

            if access_info.vector_index is not None:
                return self.codegen_state.ast_builder.add_vector_access(
                    result, access_info.vector_index)
            else:
                return result

        else:
            raise AssertionError()

    def map_linear_subscript(self, expr, type_context):
        from pymbolic.primitives import Variable
        if not isinstance(expr.aggregate, Variable):
            raise RuntimeError("linear indexing on non-variable: %s"
                    % expr)

        if expr.aggregate.name in self.kernel.arg_dict:
            arg = self.kernel.arg_dict[expr.aggregate.name]

            from loopy.kernel.data import ImageArg
            if isinstance(arg, ImageArg):
                raise RuntimeError("linear indexing is not supported on images: %s"
                        % expr)

            else:
                # GlobalArg
                if arg.offset:
                    offset = Variable(arg.offset)
                else:
                    offset = 0

                return self.make_subscript(
                        arg,
                        var(expr.aggregate.name),
                        self.rec(offset + expr.index, "i"))

        elif expr.aggregate.name in self.kernel.temporary_variables:
            raise RuntimeError("linear indexing is not supported on temporaries: %s"
                    % expr)

        else:
            raise RuntimeError(
                    "nothing known about variable '%s'" % expr.aggregate.name)

    def make_subscript(self, array, base_expr, subscript):
        return base_expr[subscript]

    def _map_integer_div_operator(self, base_func_name, op_func, expr, type_context):
        from loopy.symbolic import get_dependencies
        iname_deps = get_dependencies(expr) & self.kernel.all_inames()
        domain = self.kernel.get_inames_domain(iname_deps)

        assumption_non_param = isl.BasicSet.from_params(self.kernel.assumptions)
        assumptions, domain = isl.align_two(assumption_non_param, domain)
        domain = domain & assumptions

        num_type = self.infer_type(expr.numerator)
        den_type = self.infer_type(expr.denominator)

        if not num_type.is_integral() or not den_type.is_integral():
            raise NotImplementedError("remainder and floordiv "
                                      "for floating-point types")

        from loopy.isl_helpers import is_nonnegative
        num_nonneg = is_nonnegative(expr.numerator, domain) \
            or num_type.numpy_dtype.kind == "u"
        den_nonneg = is_nonnegative(expr.denominator, domain) \
            or den_type.numpy_dtype.kind == "u"

        result_dtype = self.infer_type(expr)
        suffix = result_dtype.numpy_dtype.type.__name__

        def seen_func(name):
            from loopy.codegen import SeenFunction
            self.codegen_state.seen_functions.add(
                    SeenFunction(
                        name, f"{name}_{suffix}",
                        (result_dtype, result_dtype),
                        (result_dtype,)))

        if den_nonneg:
            if num_nonneg:
                return op_func(
                        self.rec(expr.numerator, "i"),
                        self.rec(expr.denominator, "i"))
            else:
                seen_func(f"{base_func_name}_pos_b")
                return var(f"{base_func_name}_pos_b_{suffix}")(
                        self.rec(expr.numerator, "i"),
                        self.rec(expr.denominator, "i"))
        else:
            seen_func(base_func_name)
            return var(f"{base_func_name}_{suffix}")(
                    self.rec(expr.numerator, "i"),
                    self.rec(expr.denominator, "i"))

    def map_floor_div(self, expr, type_context):
        import operator
        return self._map_integer_div_operator(
                "loopy_floor_div", operator.floordiv, expr, type_context)

    def map_remainder(self, expr, type_context):
        tgt_dtype = self.infer_type(expr)
        if tgt_dtype.is_complex():
            raise RuntimeError("complex remainder not defined")

        import operator
        return self._map_integer_div_operator(
                "loopy_mod", operator.mod, expr, type_context)

    def map_if(self, expr, type_context):
        from loopy.types import to_loopy_type
        result_type = self.infer_type(expr)
        return type(expr)(
                self.rec(expr.condition, type_context,
                         to_loopy_type(np.bool_)),
                self.rec(expr.then, type_context, result_type),
                self.rec(expr.else_, type_context, result_type),
                )

    def map_comparison(self, expr, type_context):
        inner_type_context = dtype_to_type_context(
                self.kernel.target,
                self.infer_type(expr.left - expr.right))

        return type(expr)(
                    self.rec(expr.left, inner_type_context),
                    expr.operator,
                    self.rec(expr.right, inner_type_context))

    def map_type_cast(self, expr: TypeCast, type_context: str):
        return self.rec(expr.child, type_context, expr.type)

    def map_constant(self, expr, type_context):
        from loopy.symbolic import Literal

        if isinstance(expr, (complex, np.complexfloating)):
            real = self.rec(expr.real, type_context)
            imag = self.rec(expr.imag, type_context)
            iota = p.Variable("I" if "I" not in self.kernel.all_variable_names()
                    else "_Complex_I")
            return real + imag*iota
        elif np.isnan(expr):
            from warnings import warn
            warn("Encountered 'bare' floating point NaN value. Since NaN != NaN,"
                 " this leads to problems with cache retrieval."
                 " Consider using `pymbolic.primitives.NaN` instead of `math.nan`."
                 " The generated code will be equivalent with the added benefit"
                 " of sound pickling/unpickling of kernel objects.", stacklevel=1)
            from pymbolic.primitives import NaN
            data_type = expr.dtype.type if isinstance(expr, np.generic) else None
            return self.map_nan(NaN(data_type), type_context)
        elif np.isneginf(expr):
            return -p.Variable("INFINITY")
        elif np.isinf(expr):
            return p.Variable("INFINITY")
        elif isinstance(expr, np.generic):
            # Explicitly typed: Generated code must reflect type exactly.

            # FIXME: This assumes a 32-bit architecture.
            if isinstance(expr, np.float32):
                return Literal(repr(float(expr))+"f")

            elif isinstance(expr, np.float64):
                return Literal(repr(float(expr)))

            # Disabled for now, possibly should be a subtarget.
            # elif isinstance(expr, np.float128):
            #     return Literal(repr(expr)+"l")

            elif isinstance(expr, np.integer):
                suffix = ""
                iinfo = np.iinfo(expr)
                if iinfo.min == 0:
                    suffix += "u"
                if iinfo.max > (2**31-1):
                    suffix += "l"
                return Literal(repr(int(expr))+suffix)
            elif isinstance(expr, np.bool_):
                return Literal("true") if expr else Literal("false")
            else:
                raise LoopyError("do not know how to generate code for "
                        "constant of numpy type '%s'" % type(expr).__name__)

        elif np.isfinite(expr):
            if type_context == "f":
                return Literal(repr(float(expr))+"f")
            elif type_context == "d":
                return Literal(repr(float(expr)))
            elif type_context in ["i", "b"]:
                return int(expr)
            else:
                if is_integer(expr):
                    return int(expr)
                raise RuntimeError("don't know how to generate code "
                        "for constant '%s'" % expr)
        else:
            raise LoopyError("don't know how to generate code "
                             "for constant '%s'" % expr)

    def map_call(self, expr, type_context):
        return (
                self.codegen_state.callables_table[
                    expr.function.name].emit_call(
                        expression_to_code_mapper=self,
                    expression=expr,
                    target=self.kernel.target))

    # {{{ deal with complex-valued variables

    def map_quotient(self, expr, type_context):
        n_dtype = self.infer_type(expr.numerator).numpy_dtype
        d_dtype = self.infer_type(expr.denominator).numpy_dtype

        num = self.rec(expr.numerator, type_context)

        # analogous to ^{-1}
        denom = self.rec(expr.denominator, type_context)

        if (n_dtype.kind not in "fc"
                and d_dtype.kind not in "fc"):
            # must both be integers
            if type_context == "f":
                num = var("(float) ")(num)
                denom = var("(float) ")(denom)
            elif type_context == "d":
                num = var("(double) ")(num)
                denom = var("(double) ")(denom)

        return type(expr)(num, denom)

    def map_power(self, expr, type_context):
        tgt_dtype = self.infer_type(expr)
        base_dtype = self.infer_type(expr.base)
        exponent_dtype = self.infer_type(expr.exponent)

        from pymbolic.primitives import is_constant, is_zero
        if is_constant(expr.exponent):
            if is_zero(expr.exponent):
                return 1
            elif is_zero(expr.exponent - 1):
                return self.rec(expr.base, type_context)
            elif is_zero(expr.exponent - 2):
                return self.rec(expr.base*expr.base, type_context)

        if exponent_dtype.is_integral():
            from loopy.codegen import SeenFunction
            func_name = ("loopy_pow_"
                    f"{tgt_dtype.numpy_dtype}_{exponent_dtype.numpy_dtype}")

            self.codegen_state.seen_functions.add(
                    SeenFunction(
                        "int_pow", func_name,
                        (tgt_dtype, exponent_dtype),
                        (tgt_dtype, )))
            # FIXME: This need some more callables to be registered.
            return var(func_name)(self.rec(expr.base, type_context),
                                  self.rec(expr.exponent, type_context))
        else:
            from loopy.codegen import SeenFunction
            clbl = self.codegen_state.ast_builder.known_callables["pow"]
            clbl = clbl.with_types({0: tgt_dtype, 1: exponent_dtype},
                    self.codegen_state.callables_table)[0]
            self.codegen_state.seen_functions.add(
                    SeenFunction(
                        clbl.name, clbl.name_in_target,
                        (base_dtype, exponent_dtype),
                        (tgt_dtype,)))
            return var(clbl.name_in_target)(self.rec(expr.base, type_context),
                    self.rec(expr.exponent, type_context))

    # }}}

    def map_group_hw_index(self, expr, type_context):
        raise LoopyError("plain C does not have group hw axes")

    def map_local_hw_index(self, expr, type_context):
        raise LoopyError("plain C does not have local hw axes")

    def map_nan(self, expr, type_context):
        from loopy.types import NumpyType
        if expr.data_type is None:
            if type_context == "f":
                return p.Variable("NAN")
            elif type_context == "d":
                registry = self.codegen_state.ast_builder.target.get_dtype_registry()
                lpy_type = NumpyType(np.dtype(np.float32))
                cast = var("(%s)" % registry.dtype_to_ctype(lpy_type))
                return cast(p.Variable("NAN"))
            else:
                raise NotImplementedError("lowering NaN with type context"
                                          f" '{type_context}'.")
        else:
            if isinstance(expr.data_type(float("nan")), np.float32):
                return p.Variable("NAN")
            elif isinstance(expr.data_type(float("nan")), np.floating):
                registry = self.codegen_state.ast_builder.target.get_dtype_registry()
                lpy_type = NumpyType(np.dtype(expr.data_type))
                cast = var("(%s)" % registry.dtype_to_ctype(lpy_type))
                return cast(p.Variable("NAN"))
            elif isinstance(expr.data_type(float("nan")), np.complexfloating):
                real_dtype = np.empty(0, dtype=expr.data_type).real.dtype.type
                return self.map_constant(real_dtype("nan") + expr.data_type(1j),
                                         type_context)
            else:
                raise NotImplementedError(f"{type(self.kernel.target)} does not"
                                          f" support NaNs of type {expr.data_type}.")

# }}}


# {{{ C expression to code mapper

class CExpressionToCodeMapper(Mapper):

    # {{{ helpers

    def parenthesize_if_needed(self, s, enclosing_prec, my_prec):
        if enclosing_prec > my_prec:
            return "(%s)" % s
        else:
            return s

    def join_rec(self, joiner, iterable, prec, force_parens_around=()):
        f = joiner.join("%s" for i in iterable)
        return f % tuple(
                self.rec_with_force_parens_around(
                    i, prec, force_parens_around=force_parens_around)
                for i in iterable)

    def rec_with_force_parens_around(
            self, expr, enclosing_prec, force_parens_around=()):
        result = self.rec(expr, enclosing_prec)

        if isinstance(expr, force_parens_around):
            result = "(%s)" % result

        return result

    def join(self, joiner, iterable):
        f = joiner.join("%s" for i in iterable)
        return f % tuple(iterable)

    # }}}

    def map_constant(self, expr, prec):
        if isinstance(expr, np.generic):
            if isinstance(expr, np.integer):
                # FIXME: Add type suffixes?
                return repr(int(expr))
            elif isinstance(expr, np.float32):
                return f"{float(expr)!r}f"
            elif isinstance(expr, np.float64):
                return repr(float(expr))
            else:
                raise NotImplementedError(
                        f"unimplemented numpy-to-C conversion: {type(expr)}")
        else:
            return repr(expr)

    def map_call(self, expr, enclosing_prec):
        from pymbolic.mapper.stringifier import PREC_CALL, PREC_NONE
        from pymbolic.primitives import Variable
        if isinstance(expr.function, Variable):
            func = expr.function.name
        else:
            func = self.rec(expr.function, PREC_CALL+1)

        return self.parenthesize_if_needed(
                "{}({})".format(
                    func,
                    self.join_rec(", ", expr.parameters, PREC_NONE)),
                enclosing_prec, PREC_CALL)

    def map_common_subexpression(self, expr, prec):
        raise RuntimeError("common subexpression should have been eliminated upon "
                "entry to loopy")

    def map_variable(self, expr, enclosing_prec):
        return expr.name

    map_tagged_variable = map_variable

    def map_lookup(self, expr, enclosing_prec):
        return self.parenthesize_if_needed(
                "{}.{}".format(
                    self.rec(expr.aggregate, PREC_CALL), expr.name),
                enclosing_prec, PREC_CALL)

    def map_subscript(self, expr, enclosing_prec):
        return self.parenthesize_if_needed(
                "{}[{}]".format(
                    self.rec(expr.aggregate, PREC_CALL+1),
                    self.rec(expr.index, PREC_NONE)),
                enclosing_prec, PREC_CALL)

    def map_min(self, expr, enclosing_prec):
        what = type(expr).__name__.lower()

        children = list(expr.children)

        result = self.rec(children.pop(), PREC_NONE)
        while children:
            result = "{}({}, {})".format(what,
                        self.rec(children.pop(), PREC_NONE),
                        result)

        return result

    map_max = map_min

    def map_if(self, expr, enclosing_prec):
        from pymbolic.mapper.stringifier import PREC_CALL, PREC_NONE
        return "({} ? {} : {})".format(
                # Force parentheses around the condition to prevent compiler
                # warnings regarding precedence (e.g. with POCL 1.8/LLVM 12):
                # "warning: pocl-cache/tempfile_BYDWne.cl:96:2241: operator '?:'
                # has lower precedence than '*'; '*' will be evaluated first"
                self.rec(expr.condition, PREC_CALL),
                self.rec(expr.then, PREC_NONE),
                self.rec(expr.else_, PREC_NONE),
                )

    def map_comparison(self, expr, enclosing_prec):
        from pymbolic.mapper.stringifier import PREC_COMPARISON

        return self.parenthesize_if_needed(
                "{} {} {}".format(
                    self.rec(expr.left, PREC_COMPARISON),
                    expr.operator,
                    self.rec(expr.right, PREC_COMPARISON)),
                enclosing_prec, PREC_COMPARISON)

    def map_literal(self, expr, enclosing_prec):
        return expr.s

    def map_left_shift(self, expr, enclosing_prec):
        return self.parenthesize_if_needed(
            self.join_rec(" << ", (expr.shiftee, expr.shift), PREC_SHIFT),
            enclosing_prec, PREC_SHIFT)

    def map_right_shift(self, expr, enclosing_prec):
        return self.parenthesize_if_needed(
            self.join_rec(" >> ", (expr.shiftee, expr.shift), PREC_SHIFT),
            enclosing_prec, PREC_SHIFT)

    def map_logical_not(self, expr, enclosing_prec):
        return self.parenthesize_if_needed(
                "!" + self.rec(expr.child, PREC_UNARY),
                enclosing_prec, PREC_UNARY)

    def map_logical_and(self, expr, enclosing_prec):
        return self.parenthesize_if_needed(
                self.join_rec(" && ", expr.children, PREC_LOGICAL_AND),
                enclosing_prec, PREC_LOGICAL_AND)

    def map_logical_or(self, expr, enclosing_prec):
        mapped_children = []
        from pymbolic.primitives import LogicalAnd
        for child in expr.children:
            mapped_child = self.rec(child, PREC_LOGICAL_OR)
            # clang warns on unparenthesized && within ||
            if isinstance(child, LogicalAnd):
                mapped_child = "(%s)" % mapped_child
            mapped_children.append(mapped_child)

        result = self.join(" || ", mapped_children)
        if enclosing_prec > PREC_LOGICAL_OR:
            result = "(%s)" % result
        return result

    def map_bitwise_not(self, expr, enclosing_prec):
        return self.parenthesize_if_needed(
                "~" + self.rec(expr.child, PREC_UNARY),
                enclosing_prec, PREC_UNARY)

    def map_bitwise_and(self, expr, enclosing_prec):
        return self.parenthesize_if_needed(
                self.join_rec(" & ", expr.children, PREC_BITWISE_AND),
                enclosing_prec, PREC_BITWISE_AND)

    def map_bitwise_or(self, expr, enclosing_prec):
        return self.parenthesize_if_needed(
                self.join_rec(" | ", expr.children, PREC_BITWISE_OR),
                enclosing_prec, PREC_BITWISE_OR)

    def map_bitwise_xor(self, expr, enclosing_prec):
        return self.parenthesize_if_needed(
                self.join_rec(" ^ ", expr.children, PREC_BITWISE_XOR),
                enclosing_prec, PREC_BITWISE_XOR)

    def map_sum(self, expr, enclosing_prec):
        from pymbolic.mapper.stringifier import PREC_SUM

        return self.parenthesize_if_needed(
                self.join_rec(" + ", expr.children, PREC_SUM),
                enclosing_prec, PREC_SUM)

    multiplicative_primitives = (p.Product, p.Quotient, p.FloorDiv, p.Remainder)

    def map_product(self, expr, enclosing_prec):
        force_parens_around = (p.Quotient, p.FloorDiv, p.Remainder)

        # Spaces prevent '**z' (times dereference z), which is hard to read.
        return self.parenthesize_if_needed(
                self.join_rec(" * ", expr.children, PREC_PRODUCT,
                    force_parens_around=force_parens_around),
                enclosing_prec, PREC_PRODUCT)

    def _map_division_operator(self, operator, expr, enclosing_prec):
        num_s = self.rec_with_force_parens_around(expr.numerator, PREC_PRODUCT,
                force_parens_around=self.multiplicative_primitives)

        denom_s = self.rec_with_force_parens_around(expr.denominator, PREC_PRODUCT,
                force_parens_around=self.multiplicative_primitives)

        return self.parenthesize_if_needed(
                f"{num_s} {operator} {denom_s}",
                # Space is necessary--otherwise '/*'
                # (i.e. divide-dererference) becomes
                # start-of-comment in C.
                enclosing_prec, PREC_PRODUCT)

    def map_quotient(self, expr, enclosing_prec):
        return self._map_division_operator("/", expr, enclosing_prec)

    def map_floor_div(self, expr, enclosing_prec):
        return self._map_division_operator("/", expr, enclosing_prec)

    def map_remainder(self, expr, enclosing_prec):
        return self._map_division_operator("%", expr, enclosing_prec)

    def map_power(self, expr, enclosing_prec):
        raise RuntimeError(f"'{expr}' should have been transformed to 'Call'"
                           " expression node.")

    def map_array_literal(self, expr, enclosing_prec):
        return "{ %s }" % self.join_rec(", ", expr.children, PREC_NONE)

# }}}

# vim: fdm=marker



from __future__ import annotations


__copyright__ = "Copyright (C) 2018 Kaushik Kulkarni"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import collections
from collections.abc import Set as abc_Set
from dataclasses import dataclass, field, replace
from functools import wraps
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Mapping,
    TypeVar,
    Union,
)
from warnings import warn

from constantdict import constantdict
from typing_extensions import Concatenate, ParamSpec, Self

from pymbolic.primitives import Call, Variable

from loopy.diagnostic import DirectCallUncachedWarning, LoopyError
from loopy.kernel.function_interface import (
    CallableKernel,
    InKernelCallable,
    ScalarCallable,
)
from loopy.library.reduction import ReductionOpFunction
from loopy.symbolic import (
    ResolvedFunction,
    RuleAwareIdentityMapper,
    SubstitutionRuleMappingContext,
)


if TYPE_CHECKING:
    from loopy.kernel import LoopKernel
    from loopy.target import TargetBase
    from loopy.target.execution import ExecutorBase


__doc__ = """

.. class:: FunctionIdT

    A type for a function identifier.
    A :class:`~loopy.library.reduction.ReductionOpFunction` or a :class:`str`.

.. class:: CallablesTable

    A type alias for callables tables, mapping from :class:`FunctionIdT`
    to :class:`~loopy.InKernelCallable`

.. currentmodule:: loopy

.. autoclass:: TranslationUnit

.. currentmodule:: loopy.translation_unit

.. autoclass:: CallablesInferenceContext

.. autofunction:: make_program

.. autofunction:: check_each_kernel

.. autofunction:: for_each_kernel

.. autoclass:: TUnitOrKernelT

.. class:: P

    A :class:`typing.ParamSpec` for use in annotating :func:`for_each_kernel` and
    :func:`check_each_kernel`.
"""


# {{{ CallableResolver

def _is_a_reduction_op(expr):
    if isinstance(expr, ResolvedFunction):
        return _is_a_reduction_op(expr.function)

    return isinstance(expr, ReductionOpFunction)


class CallableResolver(RuleAwareIdentityMapper):
    """
    Resolves callables in expressions and records the names of the calls
    resolved.

    .. attribute:: known_callables

        An instance of :class:`frozenset` of the call names to be resolved.

    .. attribute:: rule_mapping_context

        An instance of :class:`loopy.symbolic.RuleMappingContext`.

    .. attribute:: calls_resolved

        A :class:`set` of calls that were resolved. Updated during an
        expression traversal.
    """
    def __init__(self, rule_mapping_context, known_callables):
        assert isinstance(known_callables, frozenset)

        super().__init__(rule_mapping_context)

        self.known_callables = known_callables

        # a record of the call names that were resolved
        self.calls_resolved = set()

    def map_call(self, expr, expn_state):
        from loopy.symbolic import parse_tagged_name

        if not _is_a_reduction_op(expr.function):
            # FIXME: We should have never used parse_tagged_name here.
            name, tag = parse_tagged_name(expr.function)

            if tag:
                raise LoopyError(f"tagged name in call: {expr.function}")

        else:
            if isinstance(expr.function, ResolvedFunction):
                name = expr.function.function
            else:
                name = expr.function

        if name in self.known_callables:
            params = tuple(self.rec(par, expn_state) for par in expr.parameters)

            # record that we resolved a call
            self.calls_resolved.add(name)

            function = expr.function

            if not isinstance(expr.function, ResolvedFunction):
                function = ResolvedFunction(expr.function)

            return Call(function, params)

        return super().map_call(expr, expn_state)

    def map_call_with_kwargs(self, expr):
        # See https://github.com/inducer/loopy/pull/323
        raise NotImplementedError

# }}}


# {{{ translation unit

FunctionIdT = Union[str, ReductionOpFunction]
ConcreteCallablesTable = constantdict[FunctionIdT, InKernelCallable]
CallablesTable = Mapping[FunctionIdT, InKernelCallable]


@dataclass(frozen=True)
class TranslationUnit:
    """
    Records the information about all the callables in a :mod:`loopy` program.

    An instance of :class:`TranslationUnit` is the object that gets lowered
    for a :class:`loopy.target.TargetBase`.


    .. attribute:: entrypoints

        A :class:`frozenset` of the names of the kernels which
        could be called from the host.

    .. attribute:: default_entrypoint

        The :class:`~loopy.LoopKernel` representing the main entrypoint
        of the program, if defined. Currently, this attribute may only be
        accessed if there is exactly one entrypoint in the translation unit.
        Will raise an error if the default entrypoint is not a
        :class:`~loopy.LoopKernel`.

    .. attribute:: callables_table

        An instance of :class:`constantdict.constantdict` mapping the function
        identifiers in a kernel to their associated instances of
        :class:`~loopy.kernel.function_interface.InKernelCallable`.

    .. attribute:: target

        An instance of :class:`loopy.target.TargetBase`.

    .. attribute:: func_id_to_in_knl_callables_mappers

        A :class:`frozenset` of functions of the signature ``(target:
        TargetBase, function_identifier: str)`` that returns an instance
        of :class:`loopy.kernel.function_interface.InKernelCallable` or *None*.

    .. automethod:: executor
    .. automethod:: __call__
    .. automethod:: copy
    .. automethod:: __getitem__
    .. automethod:: with_kernel

    .. note::

        - To create an instance of :class:`loopy.TranslationUnit`, it is
          recommended to go through :func:`loopy.make_kernel`.
        - This data structure and its attributes should be considered
          immutable, any modifications should be done through
          :meth:`~TranslationUnit.copy`.

    """

    callables_table: ConcreteCallablesTable
    target: TargetBase
    entrypoints: frozenset[str]

    def __post_init__(self):
        assert isinstance(self.entrypoints, abc_Set)
        assert isinstance(self.callables_table, constantdict)

    def copy(self, **kwargs: Any) -> Self:
        target = kwargs.pop("target", None)
        t_unit = replace(self, **kwargs)
        if target:
            from loopy.kernel import KernelState
            if max(callable_knl.subkernel.state
                   for callable_knl in self.callables_table.values()
                   if isinstance(callable_knl, CallableKernel)) > (
                            KernelState.INITIAL):
                if not isinstance(target, type(self.target)):
                    raise LoopyError("One of the kernels in the program has been "
                            "preprocessed, cannot modify target now.")

            new_callables = {}
            for func_id, clbl in t_unit.callables_table.items():
                if isinstance(clbl, CallableKernel):
                    knl = clbl.subkernel
                    knl = knl.copy(target=target)
                    clbl = clbl.copy(subkernel=knl)
                elif isinstance(clbl, ScalarCallable):
                    pass
                else:
                    raise NotImplementedError()
                new_callables[func_id] = clbl

            t_unit = replace(
                    self, callables_table=constantdict(new_callables), target=target)

        return t_unit

    def with_entrypoints(self, entrypoints: str | frozenset[str]) -> Self:
        if isinstance(entrypoints, str):
            entrypoints = frozenset([e.strip() for e in
                entrypoints.split(",")])

        assert isinstance(entrypoints, abc_Set)

        return self.copy(entrypoints=entrypoints)

    @property
    def state(self):
        """ Returns an instance of :class:`loopy.kernel.KernelState`. """
        from loopy.kernel import KernelState
        return min((callable_knl.subkernel.state
                    for callable_knl in self.callables_table.values()
                    if isinstance(callable_knl, CallableKernel)),
                   default=KernelState.INITIAL)

    def with_kernel(self, kernel: LoopKernel) -> Self:
        """
        If *self* contains a callable kernel with *kernel*'s name, replaces its
        subkernel and returns a copy of *self*. Else records a new callable
        kernel with *kernel* as its subkernel.

        :arg kernel: An instance of :class:`loopy.LoopKernel`.
        :returns: Copy of *self* with updated callable kernels.
        """
        if kernel.name in self.callables_table:
            # update the callable kernel
            new_in_knl_callable = self.callables_table[kernel.name].copy(
                    subkernel=kernel)
            return self.copy(
                callables_table=self.callables_table.set(
                                        kernel.name, new_in_knl_callable))
        else:
            # add a new callable kernel
            clbl = CallableKernel(kernel)
            return self.copy(
                callables_table=self.callables_table.set(kernel.name, clbl))

    def __getitem__(self, name) -> LoopKernel:
        """
        For the callable named *name*, return a :class:`loopy.LoopKernel`. if
        it's a :class:`~loopy.kernel.function_interface.CallableKernel`
        otherwise return the callable itself.
        """
        result = self.callables_table[name]
        if isinstance(result, CallableKernel):
            return result.subkernel
        else:
            raise ValueError("TranslationUnit.__getitem__ "
                             "can only be used for instances of LoopKernel. "
                             "Access all other callables via callables_table.")

    @property
    def default_entrypoint(self) -> LoopKernel:
        if len(self.entrypoints) == 1:
            ep_name, = self.entrypoints
            entrypoint = self[ep_name]

            from loopy import LoopKernel
            if not isinstance(entrypoint, LoopKernel):
                raise ValueError("default entrypoint is not a kernel")

            return entrypoint
        else:
            raise ValueError("TranslationUnit has multiple possible entrypoints."
                             " The default entrypoint kernel is not uniquely"
                             " determined.")

    def executor(self,
                 *args, entrypoint: str | None = None, **kwargs) -> ExecutorBase:
        """Return an object that hosts caches of compiled code for execution (i.e.
        a subclass of :class:`ExecutorBase`, specific to an execution
        environment (e.g. an OpenCL context) and a given entrypoint.

        :arg entrypoint: The name of the entrypoint callable to be called.
            Defaults to :attr:`default_entrypoint`.
            An error will result if multiple entrypoints exist and no
            entrypoint is specified.

        The variable arguments to this are target-specific. The
        :class:`PyOpenCLTarget` takes a :class:`~pyopencl.Context` or a
        :class:`~pyopencl.CommandQueue`.
        """
        if entrypoint is None:
            nentrypoints = len(self.entrypoints)
            if nentrypoints == 1:
                entrypoint, = self.entrypoints
            elif nentrypoints > 1:
                raise ValueError("TranslationUnit has multiple possible entrypoints."
                                 " The default entrypoint kernel is not uniquely"
                                 " determined. You may explicitly specify an "
                                 " entrypoint using the 'entrypoint' kwarg.")
            elif nentrypoints == 0:
                raise ValueError("TranslationUnit has no entrypoints, but"
                                 f" {len(self.callables_table)} callables."
                                 " Use TranslationUnit.with_entrypoints to"
                                 " set an entrypoint.")
            else:
                raise AssertionError
        else:
            if entrypoint not in self.entrypoints:
                raise LoopyError(f"'{entrypoint}' not in list of possible "
                        "entrypoints for the translation unit. "
                        "Maybe you want to invoke 'with_entrypoints' before "
                        "calling the translation unit?")

        return self.target.get_kernel_executor(self, *args,
                                               entrypoint=entrypoint, **kwargs)

    def __call__(self, *args, **kwargs):
        """
        Builds and calls the *entrypoint* kernel, if
        :attr:`TranslationUnit.target` is an executable target.

        :arg entrypoint: The name of the entrypoint callable to be called.
            Defaults to :attr:`default_entrypoint`.

        .. warning::

            While this was the main execution interface for loopy for many
            years (and reasonably efficient), the caches that made this so
            kept lots of expensive 'stuff' (such as OpenCL contexts) alive
            for no good reason, leading to major inefficiencies.
            See :meth:`executor` for an efficient, cached way to
            invoke kernels.
        """

        # The rationale for this is that the executor cache held long-lived
        # references to OpenCL contexts, and translation units were kept alive
        # long-term by caches, leading to many stale contexts being kept alive.
        # While attempts were made to turn those into weak references, this was
        # ultimately cumbersome and ineffective.
        #
        # In addition, the executor interface speeds up kernel invocation
        # by removing one unnecessary layer of function call.
        warn("TranslationUnit.__call__ is uncached as of 2025, "
             "meaning it will incur possibly substantial compilation cost "
             "with every invocation. Use TranslationUnit.executor to obtain "
             "an object that holds longer-lived caches.",
             DirectCallUncachedWarning, stacklevel=2)

        entrypoint = kwargs.get("entrypoint", None)

        if entrypoint is None:
            nentrypoints = len(self.entrypoints)
            if nentrypoints == 1:
                entrypoint, = self.entrypoints
            elif nentrypoints > 1:
                raise ValueError("TranslationUnit has multiple possible entrypoints."
                                 " The default entrypoint kernel is not uniquely"
                                 " determined. You may explicitly specify an "
                                 " entrypoint using the 'entrypoint' kwarg.")
            elif nentrypoints == 0:
                raise ValueError("TranslationUnit has no entrypoints, but"
                                 f" {len(self.callables_table)} callables."
                                 " Use TranslationUnit.with_entrypoints to"
                                 " set an entrypoint.")
            else:
                raise AssertionError
        else:
            if entrypoint not in self.entrypoints:
                raise LoopyError(f"'{entrypoint}' not in list of possible "
                        "entrypoints for the translation unit. "
                        "Maybe you want to invoke 'with_entrypoints' before "
                        "calling the translation unit?")

        kwargs["entrypoint"] = entrypoint

        pex = self.target.get_kernel_executor(self, *args, **kwargs)

        del kwargs["entrypoint"]

        return pex(*args, **kwargs)

    def __str__(self):
        # FIXME: do a topological sort by the call graph

        return "\n".join(
                str(clbl.subkernel)
                for _name, clbl in self.callables_table.items()
                if isinstance(clbl, CallableKernel))

    # FIXME: This is here because Firedrake expects it, for some legacy reason.
    # Without that, it would be safe to delete.
    def update_persistent_hash(self, key_hash, key_builder):
        key_builder.update_for_dataclass(key_hash, self)

# }}}


# {{{ rename resolved functions

class ResolvedFunctionRenamer(RuleAwareIdentityMapper):
    """
    Mapper to rename the resolved functions in an expression according to
    *renaming_dict*.
    """
    def __init__(self, rule_mapping_context, renaming_dict):
        super().__init__(
                rule_mapping_context)
        self.renaming_dict = renaming_dict

    def map_resolved_function(self, expr, expn_state):
        if expr.name in self.renaming_dict:
            return ResolvedFunction(self.renaming_dict[expr.name])
        else:
            return super().map_resolved_function(
                    expr, expn_state)


def rename_resolved_functions_in_a_single_kernel(kernel,
        renaming_dict):
    """
    Returns a copy of *kernel* with the instances of :class:`ResolvedFunction`
    renames according to *renaming_dict*.
    """
    from loopy.symbolic import SubstitutionRuleMappingContext
    rule_mapping_context = SubstitutionRuleMappingContext(
                kernel.substitutions, kernel.get_var_name_generator())
    resolved_function_renamer = ResolvedFunctionRenamer(rule_mapping_context,
            renaming_dict)
    return (
            rule_mapping_context.finish_kernel(
                resolved_function_renamer.map_kernel(kernel)))

# }}}


def get_reachable_resolved_callable_ids(callables, entrypoints):
    """
    Returns a :class:`frozenset` of callables ids that are resolved and
    reachable from *entrypoints*.
    """
    return frozenset().union(*(callables[e].get_called_callables(callables)
                               for e in entrypoints))


# {{{ CallablesInferenceContext

def get_all_subst_names(callables):
    """
    Returns a :class:`set` of all substitution rule names in the callable
    kernels of *callables*.

    :arg callables: A mapping from function identifiers to
        :class:`~loopy.kernel.function_interface.InKernelCallable`.
    """
    return set().union(*(set(clbl.subkernel.substitutions.keys())
                         for clbl in callables.values()
                         if isinstance(clbl, CallableKernel)))


def make_callable_name_generator(callables):
    from pytools import UniqueNameGenerator
    all_substs = get_all_subst_names(callables)
    return UniqueNameGenerator(set(callables.keys()) | all_substs)


def make_clbl_inf_ctx(callables, entrypoints):
    name_gen = make_callable_name_generator(callables)
    return CallablesInferenceContext(callables, name_gen)


@dataclass(frozen=True)
class CallablesInferenceContext:
    """
    Helper class for housekeeping a :attr:`loopy.TranslationUnit.callables_table`
    while traversing through callables of :class:`loopy.TranslationUnit`.

    .. attribute:: callables

       A mapping from the callable names to instances of
       :class:`loopy.kernel.function_interface.InKernelCallable`.

    .. attribute:: renames

       A mapping from old function identifiers to a :class:`frozenset` of new
       function identifiers.

    .. attribute:: new_entrypoints

       A :class:`frozenset` of renamed entrypoint names.

    .. automethod:: with_callable

    .. automethod:: finish_program

    .. automethod:: __getitem__
    """
    callables: Mapping[str, InKernelCallable]
    clbl_name_gen: Callable[[str], str]
    renames: Mapping[str, frozenset[str]] = field(
            default_factory=lambda: collections.defaultdict(frozenset))
    new_entrypoints: frozenset[str] = frozenset()

    def copy(self, **kwargs: Any) -> CallablesInferenceContext:
        return replace(self, **kwargs)

    def with_callable(self, old_function_id, new_clbl,
                      is_entrypoint=False):
        """
        Updates the callable referred by *function_id*'s in *self*'s namespace
        to *new_clbl*.

        :arg old_function_id: An instance of :class:`pymbolic.primitives.Variable` or
            :class:`loopy.library.reduction.ReductionOpFunction`.

        :arg new_clbl: An instance of
            :class:`loopy.kernel.function_interface.InKernelCallable`.

        :returns: ``(new_self, new_function_id)`` is a copy of *self* with
            *new_clbl* in its namespace. *new_clbl* would be referred by
            *new_function_id* in *new_self*'s namespace.
        """

        assert isinstance(old_function_id, (str, Variable, ReductionOpFunction))

        if isinstance(old_function_id, Variable):
            old_function_id = old_function_id.name

        renames = collections.defaultdict(frozenset, self.renames)

        # if the callable already exists => return the function
        # identifier corresponding to that callable.
        for func_id, clbl in self.callables.items():
            if clbl == new_clbl:
                renames[old_function_id] |= frozenset([func_id])
                if isinstance(func_id, str):
                    new_entrypoints = self.new_entrypoints
                    if is_entrypoint:
                        new_entrypoints |= frozenset([func_id])
                    return (self.copy(renames=renames,
                                      new_entrypoints=new_entrypoints),
                            Variable(func_id),)
                else:
                    assert not is_entrypoint
                    assert isinstance(func_id, ReductionOpFunction)
                    return (self.copy(renames=renames),
                            func_id)

        # {{{ handle ReductionOpFunction

        if isinstance(old_function_id, ReductionOpFunction):
            # FIXME: Check if we have 2 ArgMax functions
            # with different types in the same kernel the generated code
            # does not mess up the types.
            assert not is_entrypoint
            unique_function_id = old_function_id.copy()
            updated_callables = self.callables.copy()
            updated_callables[unique_function_id] = new_clbl
            renames[old_function_id] |= frozenset([unique_function_id])

            return (self.copy(callables=updated_callables,
                              renames=renames),
                    unique_function_id)

        # }}}

        # must allocate a new clbl in the namespace => find a unique id for it
        unique_function_id = self.clbl_name_gen(old_function_id)

        updated_callables = dict(self.callables)
        updated_callables[unique_function_id] = new_clbl
        renames[old_function_id] |= frozenset([unique_function_id])

        new_entrypoints = self.new_entrypoints
        if is_entrypoint:
            new_entrypoints |= frozenset([unique_function_id])

        return (self.copy(renames=renames,
                          callables=updated_callables,
                          new_entrypoints=new_entrypoints),
                Variable(unique_function_id))

    def finish_program(self, program):
        """
        Returns a copy of *program* with rollback renaming of the callables
        done whenever possible.

        For example: If all the ``sin`` function ids diverged as
        ``sin_0``, ``sin_1``, then all the renaming is done such that one of
        the flavors of the callable is renamed back to ``sin``.
        """
        # FIXME: Generalize this if an inference happens over a proper subgraph
        # of the callgraph (the following assert should be removed)
        assert len(self.new_entrypoints) == len(program.entrypoints)

        # {{{ get all the callables reachable from the new entrypoints.

        # get the names of all callables reachable from the new entrypoints
        new_callable_ids = get_reachable_resolved_callable_ids(self.callables,
                                                               self.new_entrypoints)

        # get the history of function ids from the performed renames:
        history = {}
        for old_func_id, new_func_ids in self.renames.items():
            for new_func_id in new_func_ids:
                if new_func_id in (new_callable_ids | self.new_entrypoints):
                    history[new_func_id] = old_func_id

        # }}}

        # {{{ preserve the entrypoints of *program*

        # If there are any callees having old entrypoint names => mark them for
        # renaming
        callees_with_old_entrypoint_names = ((program.entrypoints & new_callable_ids)
                                             - self.new_entrypoints)

        todo_renames = {}
        new_callables = dict(program.callables_table)

        for c in callees_with_old_entrypoint_names:
            todo_renames[c] = self.clbl_name_gen(c)

        for e in self.new_entrypoints:
            # note renames to "rollback" the renaming of entrypoints
            todo_renames[e] = history[e]
            assert todo_renames[e] in program.entrypoints

        # }}}

        # try to roll back the names as much as possible
        for new_id in new_callable_ids:
            old_func_id = history[new_id]
            if (isinstance(old_func_id, str)
                    and old_func_id not in set(todo_renames.values())):
                todo_renames[new_id] = old_func_id

        # {{{ perform the renames form todo_renames

        for func_id in (new_callable_ids | self.new_entrypoints):
            clbl = self.callables[func_id]
            if func_id in todo_renames:
                assert history[func_id] == todo_renames[func_id]
                func_id = todo_renames[func_id]
            if isinstance(clbl, CallableKernel):
                subknl = clbl.subkernel.copy(name=func_id)
                subknl = rename_resolved_functions_in_a_single_kernel(subknl,
                                                                      todo_renames)

                clbl = clbl.copy(subkernel=subknl)

            new_callables[func_id] = clbl

        # }}}

        return program.copy(callables_table=constantdict(new_callables))

    def __getitem__(self, name):
        result = self.callables[name]
        return result

# }}}


TUnitOrKernelT = TypeVar("TUnitOrKernelT", "LoopKernel", TranslationUnit)


# {{{ helper functions

def make_program(kernel: LoopKernel) -> TranslationUnit:
    """
    Returns an instance of :class:`loopy.TranslationUnit` with *kernel* as the only
    callable kernel.
    """

    return TranslationUnit(
            callables_table=constantdict({
                kernel.name: CallableKernel(kernel)}),
            target=kernel.target,
            entrypoints=frozenset())


P = ParamSpec("P")


def check_each_kernel(
            check: Callable[Concatenate[LoopKernel, P], None]
        ) -> Callable[Concatenate[TranslationUnit, P], None]:
    def _collective_check(
                t_unit_or_kernel: TranslationUnit | LoopKernel, /,
                *args: P.args,
                **kwargs: P.kwargs
            ) -> None:
        from loopy import LoopKernel
        if isinstance(t_unit_or_kernel, TranslationUnit):
            for clbl in t_unit_or_kernel.callables_table.values():
                if isinstance(clbl, CallableKernel):
                    check(clbl.subkernel, *args, **kwargs)
                elif isinstance(clbl, ScalarCallable):
                    pass
                else:
                    raise NotImplementedError(f"{type(clbl)}")
        elif isinstance(t_unit_or_kernel, LoopKernel):
            check(t_unit_or_kernel, *args, **kwargs)
        else:
            raise TypeError("expected LoopKernel or TranslationUnit")

    return wraps(check)(_collective_check)


def for_each_kernel(
            transform: Callable[Concatenate[LoopKernel, P], LoopKernel]
        ) -> Callable[Concatenate[TUnitOrKernelT, P], TUnitOrKernelT]:
    """
    Function wrapper for transformations of the type ``transform(kernel:
    LoopKernel, *args, **kwargs) -> LoopKernel``. Returns a function that would
    apply *transform* to all callable kernels in a :class:`loopy.TranslationUnit`.
    """
    def _collective_transform(
                t_unit_or_kernel: TUnitOrKernelT, /,
                *args: P.args,
                **kwargs: P.kwargs
            ) -> TUnitOrKernelT:
        from loopy import LoopKernel
        if isinstance(t_unit_or_kernel, TranslationUnit):
            t_unit = t_unit_or_kernel
            new_callables = {}
            for func_id, clbl in t_unit.callables_table.items():
                if isinstance(clbl, CallableKernel):
                    new_subkernel = transform(clbl.subkernel, *args, **kwargs)
                    clbl = clbl.copy(subkernel=new_subkernel)
                elif isinstance(clbl, ScalarCallable):
                    pass
                else:
                    raise NotImplementedError(f"{type(clbl)}")

                new_callables[func_id] = clbl

            return t_unit.copy(callables_table=constantdict(new_callables))
        elif isinstance(t_unit_or_kernel, LoopKernel):
            kernel = t_unit_or_kernel
            return transform(kernel, *args, **kwargs)
        else:
            raise TypeError("expected LoopKernel or TranslationUnit")

    return wraps(transform)(_collective_transform)


def add_callable_to_table(callables_table, clbl_id, clbl):
    """
    Returns a tuple ``new_clbl_id, new_callables_table`` where
    *new_callables_table* is a copy of *callables_table* with *clbl* in its
    namespace. *clbl* is referred to in *new_callables_table*'s namespace by
    *new_clbl_id*.

    :arg clbl_id: An instance of :class:`str` or
        :class:`~loopy.library.reduction.ReductionOpFunction` based on which
        the unique identifier, *new_clbl_id* , is to be chosen.
    """
    from loopy.kernel.function_interface import InKernelCallable
    assert isinstance(clbl, InKernelCallable)

    for i, c in callables_table.items():
        if c == clbl:
            return i, callables_table

    if isinstance(clbl_id, ReductionOpFunction):
        new_clbl_id = clbl_id
    else:
        assert isinstance(clbl_id, str)
        ung = make_callable_name_generator(callables_table)
        new_clbl_id = ung(clbl_id)

    new_callables_table = callables_table.copy()
    new_callables_table[new_clbl_id] = clbl.with_name(new_clbl_id)

    return new_clbl_id, new_callables_table

# }}}


# {{{ resolve_callables

def resolve_callables(t_unit: TranslationUnit) -> TranslationUnit:
    """
    Returns a :class:`TranslationUnit` with known :class:`pymbolic.primitives.Call`
    expression nodes converted to :class:`loopy.symbolic.ResolvedFunction`.
    """
    from loopy.check import validate_kernel_call_sites
    from loopy.kernel import KernelState
    from loopy.library.function import get_loopy_callables

    if t_unit.state >= KernelState.CALLS_RESOLVED:
        # program's callables have been resolved
        return t_unit

    # get registered callables
    known_callables = dict(t_unit.callables_table)
    # get target specific callables
    known_callables.update(t_unit.target.get_device_ast_builder().known_callables)
    # get loopy specific callables
    known_callables.update(get_loopy_callables())

    callables_table: dict[FunctionIdT, InKernelCallable] = {}

    # callables: name of the calls seen in the program
    callables = {name for name, clbl in t_unit.callables_table.items()
                 if isinstance(clbl, CallableKernel)}

    while callables:
        clbl_name = callables.pop()
        clbl = known_callables[clbl_name]

        if isinstance(clbl, CallableKernel):
            knl = clbl.subkernel

            rule_mapping_context = SubstitutionRuleMappingContext(
                    knl.substitutions, knl.get_var_name_generator())
            clbl_resolver = CallableResolver(rule_mapping_context,
                                             frozenset(known_callables))
            knl = rule_mapping_context.finish_kernel(clbl_resolver.map_kernel(knl))
            knl = knl.copy(state=KernelState.CALLS_RESOLVED)

            # add the updated callable kernel to the table
            callables_table[clbl_name] = clbl.copy(subkernel=knl)

            # note the resolved callable for traversal
            callables.update(clbl_resolver.calls_resolved - set(callables_table))
        elif isinstance(clbl, ScalarCallable):
            # nothing to resolve within a scalar callable
            callables_table[clbl_name] = clbl
        else:
            raise NotImplementedError(f"{type(clbl)}")

    t_unit = t_unit.copy(callables_table=constantdict(callables_table))

    validate_kernel_call_sites(t_unit)

    return t_unit

# }}}


# vim: foldmethod=marker
